from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pylav.players.query.local_files import LocalFile


def is_url(uri: str | LocalFile) -> bool:
    """Check if the given uri is a url."""
    return f"{uri}".startswith(("https://", "http://", "s3://", "s3a://", "s3n://"))



from __future__ import annotations

from typing import TYPE_CHECKING

from pylav.events.base import PyLavEvent
from pylav.nodes.api.responses.plugins import SegmentSkipped, SegmentsLoaded

__all__ = ("SegmentSkippedEvent", "SegmentsLoadedEvent")

if TYPE_CHECKING:
    from pylav.nodes.node import Node
    from pylav.players.player import Player


class SegmentSkippedEvent(PyLavEvent):
    """This event is dispatched when a segment is skipped.

    Event can be listened to by adding a listener with the name `pylav_segment_skipped_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that skipped the segment.
    segment: :class:`Segment`
        The segment that was skipped.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`SegmentSkipped`
        The raw event object.

    """

    __slots__ = ("player", "segment", "node", "event")

    def __init__(self, player: Player, node: Node, event_object: SegmentSkipped) -> None:
        self.player = player
        self.segment = event_object.segment
        self.node = node
        self.event = event_object


class SegmentsLoadedEvent(PyLavEvent):
    """This event is dispatched when segments are loaded.

    Event can be listened to by adding a listener with the name `pylav_segments_loaded_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that loaded the segments.
    segments: :class:`list[Segment]`
        The segments that were loaded.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`SegmentsLoaded`
        The raw event object.

    """

    __slots__ = ("player", "segments", "node", "event")

    def __init__(self, player: Player, node: Node, event_object: SegmentsLoaded) -> None:
        self.player = player
        self.segments = event_object.segments
        self.node = node
        self.event = event_object



from __future__ import annotations

from piccolo.columns import M2M, BigInt, LazyTableReference, Text
from piccolo.columns.indexes import IndexMethod
from piccolo.table import Table

from pylav.storage.database.tables.misc import DATABASE_ENGINE


class PlaylistRow(Table, db=DATABASE_ENGINE, tablename="playlist"):
    id = BigInt(primary_key=True)
    scope = BigInt(null=True, default=None, index=True)
    author = BigInt(null=True, default=None, index=True)
    name = Text(null=True, default=None, index=True, index_method=IndexMethod.gin)
    url = Text(null=True, default=None, index=True)
    tracks = M2M(LazyTableReference("TrackToPlaylists", module_path="pylav.storage.database.tables.m2m"))



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

import discord
from redbot.core.i18n import Translator

from pylav.core.context import PyLavContext
from pylav.extension.red.ui.buttons.generic import CloseButton, NavigateButton, RefreshButton
from pylav.extension.red.ui.buttons.player import DisconnectButton
from pylav.extension.red.ui.buttons.queue import (
    DecreaseVolumeButton,
    EmptyQueueButton,
    EnqueueButton,
    IncreaseVolumeButton,
    PauseTrackButton,
    PlayNowFromQueueButton,
    PreviousTrackButton,
    QueueHistoryButton,
    RemoveFromQueueButton,
    ResumeTrackButton,
    ShuffleButton,
    SkipTrackButton,
    StopTrackButton,
    ToggleRepeatButton,
    ToggleRepeatQueueButton,
)
from pylav.extension.red.ui.menus.generic import BaseMenu
from pylav.extension.red.ui.selectors.queue import QueueSelectTrack
from pylav.extension.red.ui.sources.queue import QueuePickerSource, QueueSource
from pylav.extension.red.utils.decorators import is_dj_logic
from pylav.type_hints.bot import DISCORD_BOT_TYPE, DISCORD_COG_TYPE, DISCORD_INTERACTION_TYPE

_ = Translator("PyLav", Path(__file__))

if TYPE_CHECKING:
    from pylav.players.player import Player


class QueueMenu(BaseMenu):
    _source: QueueSource

    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        bot: DISCORD_BOT_TYPE,
        source: QueueSource,
        original_author: discord.abc.User,
        *,
        delete_after_timeout: bool = True,
        timeout: int = 600,
        message: discord.Message = None,
        starting_page: int = 0,
        history: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            cog=cog,
            bot=bot,
            source=source,
            delete_after_timeout=delete_after_timeout,
            timeout=timeout,
            message=message,
            starting_page=starting_page,
            **kwargs,
        )
        self.author = original_author
        self.is_history = history
        self.forward_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK RIGHT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}",
            direction=1,
            row=0,
            cog=cog,
        )
        self.backward_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK LEFT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}",
            direction=-1,
            row=0,
            cog=cog,
        )
        self.first_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK LEFT-POINTING DOUBLE TRIANGLE}",
            direction=0,
            row=0,
            cog=cog,
        )
        self.last_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE}",
            direction=self.source.get_max_pages,
            row=0,
            cog=cog,
        )
        self.refresh_button = RefreshButton(
            style=discord.ButtonStyle.grey,
            row=0,
            cog=cog,
        )

        self.queue_disconnect = DisconnectButton(
            style=discord.ButtonStyle.red,
            row=1,
            cog=cog,
        )
        self.repeat_queue_button_on = ToggleRepeatQueueButton(
            style=discord.ButtonStyle.blurple,
            row=1,
            cog=cog,
        )
        self.repeat_button_on = ToggleRepeatButton(
            style=discord.ButtonStyle.blurple,
            row=1,
            cog=cog,
        )
        self.repeat_button_off = ToggleRepeatButton(
            style=discord.ButtonStyle.grey,
            row=1,
            cog=cog,
        )
        self.show_history_button = QueueHistoryButton(
            style=discord.ButtonStyle.grey,
            row=1,
            cog=cog,
        )

        self.close_button = CloseButton(
            style=discord.ButtonStyle.red,
            row=1,
            cog=cog,
        )
        self.clear_queue_button = EmptyQueueButton(style=discord.ButtonStyle.red, row=1, cog=cog)

        self.previous_track_button = PreviousTrackButton(
            style=discord.ButtonStyle.grey,
            row=2,
            cog=cog,
        )
        self.stop_button = StopTrackButton(
            style=discord.ButtonStyle.grey,
            row=2,
            cog=cog,
        )
        self.paused_button = PauseTrackButton(
            style=discord.ButtonStyle.blurple,
            row=2,
            cog=cog,
        )
        self.resume_button = ResumeTrackButton(
            style=discord.ButtonStyle.blurple,
            row=2,
            cog=cog,
        )
        self.skip_button = SkipTrackButton(
            style=discord.ButtonStyle.grey,
            row=2,
            cog=cog,
        )
        self.shuffle_button = ShuffleButton(
            style=discord.ButtonStyle.grey,
            row=2,
            cog=cog,
        )

        self.decrease_volume_button = DecreaseVolumeButton(
            style=discord.ButtonStyle.grey,
            row=3,
            cog=cog,
        )
        self.increase_volume_button = IncreaseVolumeButton(
            style=discord.ButtonStyle.grey,
            row=3,
            cog=cog,
        )

        self.enqueue_button = EnqueueButton(
            cog=cog,
            style=discord.ButtonStyle.green,
            row=3,
        )
        self.remove_from_queue_button = RemoveFromQueueButton(
            cog=cog,
            style=discord.ButtonStyle.red,
            row=3,
        )
        self.play_now_button = PlayNowFromQueueButton(
            cog=cog,
            style=discord.ButtonStyle.blurple,
            row=3,
        )

    async def prepare(self):
        self.clear_items()
        max_pages = self.source.get_max_pages()
        is_dj = await is_dj_logic(self.ctx)

        if (not self.is_history) and is_dj is True:
            self.add_item(self.close_button)
            self.add_item(self.queue_disconnect)
            self.add_item(self.clear_queue_button)

        self.add_item(self.first_button)
        self.add_item(self.backward_button)
        self.add_item(self.forward_button)
        self.add_item(self.last_button)
        self.add_item(self.refresh_button)

        if not self.is_history:
            self.repeat_button_on.disabled = False
            self.repeat_button_off.disabled = False
            self.repeat_queue_button_on.disabled = False
            self.clear_queue_button.disabled = False
            self.show_history_button.disabled = False

        self.forward_button.disabled = False
        self.backward_button.disabled = False
        self.first_button.disabled = False
        self.last_button.disabled = False
        self.refresh_button.disabled = False

        if max_pages == 1:
            self.forward_button.disabled = True
            self.backward_button.disabled = True
            self.first_button.disabled = True
            self.last_button.disabled = True
        elif max_pages == 2:
            self.first_button.disabled = True
            self.last_button.disabled = True
        if self.is_history or is_dj is False:
            return

        self.previous_track_button.disabled = False
        self.paused_button.disabled = False
        self.resume_button.disabled = False
        self.stop_button.disabled = False
        self.skip_button.disabled = False
        self.shuffle_button.disabled = False

        self.decrease_volume_button.disabled = False
        self.increase_volume_button.disabled = False
        self.enqueue_button.disabled = False
        self.remove_from_queue_button.disabled = False
        self.play_now_button.disabled = False

        self.add_item(self.previous_track_button)
        self.add_item(self.stop_button)

        if (player := self.cog.pylav.get_player(self.source.guild_id)) and is_dj is not False:
            await self._player_and_dj(player)
        else:
            await self._no_player_or_no_dj()
        self.add_item(self.skip_button)
        self.add_item(self.shuffle_button)
        self.add_item(self.decrease_volume_button)
        self.add_item(self.increase_volume_button)
        self.add_item(self.enqueue_button)
        self.add_item(self.remove_from_queue_button)
        self.add_item(self.play_now_button)

    async def _player_and_dj(self, player: Player):
        self.queue_disconnect.disabled = False
        if player.paused:
            self.add_item(self.resume_button)
        else:
            self.add_item(self.paused_button)
        if player.queue.empty():
            self.shuffle_button.disabled = True
            self.remove_from_queue_button.disabled = True
            self.play_now_button.disabled = True
            self.clear_queue_button.disabled = True
        if not player.current:
            self.stop_button.disabled = True
            self.shuffle_button.disabled = True
            self.previous_track_button.disabled = True
            self.decrease_volume_button.disabled = True
            self.increase_volume_button.disabled = True
        if player.history.empty():
            self.previous_track_button.disabled = True
            self.show_history_button.disabled = True
        else:
            self.add_item(self.show_history_button)
        if await player.config.fetch_repeat_current():
            self.add_item(self.repeat_button_on)
        elif await player.config.fetch_repeat_queue():
            self.add_item(self.repeat_queue_button_on)
        else:
            self.add_item(self.repeat_button_off)

    async def _no_player_or_no_dj(self):
        self.queue_disconnect.disabled = True
        self.forward_button.disabled = True
        self.backward_button.disabled = True
        self.first_button.disabled = True
        self.last_button.disabled = True
        self.stop_button.disabled = True
        self.skip_button.disabled = True
        self.previous_track_button.disabled = True
        self.repeat_button_off.disabled = True
        self.show_history_button.disabled = True
        self.shuffle_button.disabled = True
        self.decrease_volume_button.disabled = True
        self.increase_volume_button.disabled = True
        self.resume_button.disabled = True
        self.repeat_button_on.disabled = True
        self.enqueue_button.disabled = True
        self.remove_from_queue_button.disabled = True
        self.play_now_button.disabled = True
        self.repeat_queue_button_on.disabled = True
        self.clear_queue_button.disabled = True
        self.add_item(self.resume_button)
        self.add_item(self.repeat_button_off)

    @property
    def source(self) -> QueueSource:
        return self._source

    async def start(self, ctx: PyLavContext | DISCORD_INTERACTION_TYPE):
        if isinstance(ctx, discord.Interaction):
            ctx = await self.cog.bot.get_context(ctx)
        if ctx.interaction and not ctx.interaction.response.is_done():
            await ctx.defer(ephemeral=True)

        self.ctx = ctx
        await self.send_initial_message(ctx)


class QueuePickerMenu(BaseMenu):
    _source: QueuePickerSource

    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        bot: DISCORD_BOT_TYPE,
        source: QueuePickerSource,
        original_author: discord.abc.User,
        *,
        delete_after_timeout: bool = True,
        timeout: int = 120,
        message: discord.Message = None,
        starting_page: int = 0,
        menu_type: Literal["remove", "play"],
        **kwargs: Any,
    ) -> None:
        super().__init__(
            cog=cog,
            bot=bot,
            source=source,
            delete_after_timeout=delete_after_timeout,
            timeout=timeout,
            message=message,
            starting_page=starting_page,
            **kwargs,
        )
        self.author = original_author
        self.menu_type = menu_type
        self.forward_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK RIGHT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}",
            direction=1,
            row=4,
            cog=cog,
        )
        self.backward_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK LEFT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}",
            direction=-1,
            row=4,
            cog=cog,
        )
        self.first_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK LEFT-POINTING DOUBLE TRIANGLE}",
            direction=0,
            row=4,
            cog=cog,
        )
        self.last_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE}",
            direction=self.source.get_max_pages,
            row=4,
            cog=cog,
        )
        self.close_button = CloseButton(
            style=discord.ButtonStyle.red,
            row=4,
            cog=cog,
        )
        self.refresh_button = RefreshButton(
            style=discord.ButtonStyle.grey,
            row=4,
            cog=cog,
        )
        self.select_view: QueueSelectTrack | None = None
        self.add_item(self.first_button)
        self.add_item(self.backward_button)
        self.add_item(self.forward_button)
        self.add_item(self.last_button)

    @property
    def source(self) -> QueuePickerSource:
        return self._source

    async def start(self, ctx: PyLavContext | DISCORD_INTERACTION_TYPE):
        if isinstance(ctx, discord.Interaction):
            ctx = await self.cog.bot.get_context(ctx)
        if ctx.interaction and not ctx.interaction.response.is_done():
            await ctx.defer(ephemeral=True)
        self.ctx = ctx
        await self.send_initial_message(ctx)

    async def send_initial_message(self, ctx: PyLavContext | DISCORD_INTERACTION_TYPE):
        await self._source.get_page(0)
        self.ctx = ctx
        kwargs = await self.source.format_page(self, [])
        await self.prepare()
        self.message = await ctx.send(view=self, ephemeral=True, **kwargs)
        return self.message

    async def show_page(self, page_number: int, interaction: DISCORD_INTERACTION_TYPE):
        await self._source.get_page(page_number)
        await self.prepare()
        self.current_page = page_number
        if self.message is not None:
            if not interaction.response.is_done():
                await interaction.response.pong()
            await self.message.edit(view=self)
        elif not interaction.response.is_done():
            await interaction.response.edit_message(view=self)

    async def prepare(self):
        self.clear_items()
        max_pages = self.source.get_max_pages()
        self.forward_button.disabled = False
        self.backward_button.disabled = False
        self.first_button.disabled = False
        self.last_button.disabled = False
        if max_pages == 1:
            self.forward_button.disabled = True
            self.backward_button.disabled = True
            self.first_button.disabled = True
            self.last_button.disabled = True
        elif max_pages == 2:
            self.first_button.disabled = True
            self.last_button.disabled = True
        self.add_item(self.close_button)
        self.add_item(self.first_button)
        self.add_item(self.backward_button)
        self.add_item(self.forward_button)
        self.add_item(self.last_button)
        if self.source.select_options:
            options = self.source.select_options
            if self.menu_type == "remove":
                title = _("Select a track to remove.")
            else:
                title = _("Select a track to play now.")
            self.remove_item(self.select_view)
            self.select_view = QueueSelectTrack(
                options=options,
                cog=self.cog,
                placeholder=title,
                interaction_type=self.menu_type,
                mapping=self.source.select_mapping,
            )
            self.add_item(self.select_view)
        if self.select_view and not self.source.select_options:
            self.remove_item(self.select_view)
            self.select_view = None



"""
This module contains various utilities for working with asyncio taken from
https://github.com/Cog-Creators/Red-DiscordBot/blob/V3/develop/redbot/core/utils/predicates.py
https://github.com/Cog-Creators/Red-DiscordBot/blob/V3/develop/redbot/core/utils/__init__.py

Original license can be found at:
https://github.com/Cog-Creators/Red-DiscordBot/blob/V3/develop/LICENSE
"""
from __future__ import annotations

import asyncio
import re
from collections.abc import (
    AsyncIterable,
    AsyncIterator,
    Awaitable,
    Callable,
    Coroutine,
    Generator,
    Iterable,
    Iterator,
    Sequence,
)
from itertools import chain
from re import Pattern
from typing import Any, TypeVar, cast

import discord
from discord.ext import commands

from pylav.constants.regex import DISCORD_CHANNEL_MENTION, DISCORD_ID, DISCORD_ROLE_MENTION, DISCORD_USER_MENTION

_T = TypeVar("_T")
_S = TypeVar("_S")

# https://github.com/PyCQA/pylint/issues/2717


# Benchmarked to be the fastest method.
def deduplicate_iterables(*iterables: Iterable[_T]) -> list[_T]:
    """
    Returns a list of all unique items in ``iterables``, in the order they
    were first encountered.
    """
    # dict insertion order is guaranteed to be preserved in 3.6+
    return list(dict.fromkeys(chain.from_iterable(iterables)))


# https://github.com/PyCQA/pylint/issues/2717
class AsyncFilter(AsyncIterator[_T], Awaitable[list[_T]]):  # pylint: disable=duplicate-bases
    """Class returned by `async_filter`. See that function for details.

    We do not recommend instantiating this class directly.
    """

    __slots__ = ("__func", "__iterable", "__generator_instance")

    def __init__(
        self,
        func: Callable[[_T], bool | Awaitable[bool]],
        iterable: AsyncIterable[_T] | Iterable[_T],
    ) -> None:
        self.__func: Callable[[_T], bool | Awaitable[bool]] = func
        self.__iterable: AsyncIterable[_T] | Iterable[_T] = iterable

        # We assign the generator strategy based on the arguments' types
        if isinstance(iterable, AsyncIterable):
            if asyncio.iscoroutinefunction(func):
                self.__generator_instance = self.__async_generator_async_pred()
            else:
                self.__generator_instance = self.__async_generator_sync_pred()
        elif asyncio.iscoroutinefunction(func):
            self.__generator_instance = self.__sync_generator_async_pred()
        else:
            raise TypeError("Must be either an async predicate, an async iterable, or both")

    async def __sync_generator_async_pred(self) -> AsyncIterator[_T]:
        for item in self.__iterable:
            if await self.__func(item):
                yield item

    async def __async_generator_sync_pred(self) -> AsyncIterator[_T]:
        async for item in self.__iterable:
            if self.__func(item):
                yield item

    async def __async_generator_async_pred(self) -> AsyncIterator[_T]:
        async for item in self.__iterable:
            if await self.__func(item):
                yield item

    async def __flatten(self) -> list[_T]:
        return [item async for item in self]

    def __aiter__(self) -> AsyncFilter[_T]:
        return self

    def __await__(self) -> Generator[Any, None, list[_T]]:
        # Simply return the generator filled into a list
        return self.__flatten().__await__()

    def __anext__(self) -> Coroutine[Any, Any, _T]:
        # This will use the generator strategy set in __init__
        return self.__generator_instance.__anext__()


def async_filter(
    func: Callable[[_T], bool | Awaitable[bool]],
    iterable: AsyncIterable[_T] | Iterable[_T],
) -> AsyncFilter[_T]:
    """Filter an (optionally async) iterable with an (optionally async) predicate.

    At least one of the arguments must be async.

    Parameters
    ----------
    func : Callable[[T], Union[bool, Awaitable[bool]]]
        A function or coroutine function which takes one item of ``iterable``
        as an argument, and returns ``True`` or ``False``.
    iterable : Union[AsyncIterable[_T], Iterable[_T]]
        An iterable or async iterable which is to be filtered.

    Raises
    ------
    TypeError
        If neither of the arguments are async.

    Returns
    -------
    AsyncFilter[T]
        An object which can either be awaited to yield a list of the filtered
        items, or can also act as an async iterator to yield items one by one.

    """
    return AsyncFilter(func, iterable)


async def async_enumerate(async_iterable: AsyncIterable[_T], start: int = 0) -> AsyncIterator[tuple[int, _T]]:
    """Async iterable version of `enumerate`.

    Parameters
    ----------
    async_iterable : AsyncIterable[T]
        The iterable to enumerate.
    start : int
        The index to start from. Defaults to 0.

    Returns
    -------
    AsyncIterator[Tuple[int, T]]
        An async iterator of tuples in the form of ``(index, item)``.

    """
    async for item in async_iterable:
        yield start, item
        start += 1


async def _sem_wrapper(sem: asyncio.Semaphore, task: Awaitable[_T] | asyncio.Future) -> _T:
    async with sem:
        return await task


def bounded_gather_iter(
    *coros_or_futures: Awaitable[_T] | asyncio.Future, limit: int = 4, semaphore: asyncio.Semaphore | None = None
) -> Iterator[Awaitable[Any]]:
    """
    An iterator that returns tasks as they are ready, but limits the
    number of tasks running at a time.

    Parameters
    ----------
    *coros_or_futures
        The awaitables to run in a bounded concurrent fashion.
    limit : Optional[`int`]
        The maximum number of concurrent tasks. Used when no ``semaphore``
        is passed.
    semaphore : Optional[:class:`asyncio.Semaphore`]
        The semaphore to use for bounding tasks. If `None`, create one
        using ``loop`` and ``limit``.

    Raises
    ------
    TypeError
        When invalid parameters are passed
    """
    loop = asyncio.get_running_loop()

    if semaphore is None:
        if not isinstance(limit, int) or limit <= 0:
            raise TypeError("limit must be an int > 0")

        semaphore = asyncio.Semaphore(limit)

    pending = []

    for cof in coros_or_futures:
        # noinspection PyProtectedMember
        if asyncio.isfuture(cof) and cof._loop is not loop:
            raise ValueError("futures are tied to different event loops")

        cof = _sem_wrapper(semaphore, cof)
        pending.append(cof)

    return asyncio.as_completed(pending)


def bounded_gather(
    *coros_or_futures,
    return_exceptions: bool = False,
    limit: int = 4,
    semaphore: asyncio.Semaphore | None = None,
) -> Awaitable[list[Any]]:
    """
    A semaphore-bounded wrapper to :meth:`asyncio.gather`.

    Parameters
    ----------
    *coros_or_futures
        The awaitables to run in a bounded concurrent fashion.
    return_exceptions : bool
        If true, gather exceptions in the result list instead of raising.
    limit : Optional[`int`]
        The maximum number of concurrent tasks. Used when no ``semaphore``
        is passed.
    semaphore : Optional[:class:`asyncio.Semaphore`]
        The semaphore to use for bounding tasks. If `None`, create one
        using ``loop`` and ``limit``.

    Raises
    ------
    TypeError
        When invalid parameters are passed
    """
    asyncio.get_running_loop()

    if semaphore is None:
        if not isinstance(limit, int) or limit <= 0:
            raise TypeError("limit must be an int > 0")

        semaphore = asyncio.Semaphore(limit)

    tasks = (_sem_wrapper(semaphore, task) for task in coros_or_futures)

    return asyncio.gather(*tasks, return_exceptions=return_exceptions)


class AsyncIter(AsyncIterator[_T], Awaitable[list[_T]]):  # pylint: disable=duplicate-bases
    """Asynchronous iterator yielding items from ``iterable``
    that sleeps for ``delay`` seconds every ``steps`` items.

    Parameters
    ----------
    iterable: Iterable
        The iterable to make async.
    delay: Union[float, int]
        The amount of time in seconds to sleep.
    steps: int
        The number of iterations between sleeps.

    Raises
    ------
    ValueError
        When ``steps`` is lower than 1.

    Examples
    --------
    >>> from pylav.utils.vendor.redbot import AsyncIter
    >>> async for value in AsyncIter(range(3)):
    ...     print(value)
    0
    1
    2

    """

    __slots__ = ("_delay", "_iterator", "_i", "_steps", "_map")

    def __init__(self, iterable: Iterable[_T], delay: float | int = 0, steps: int = 100) -> None:
        if steps < 1:
            raise ValueError("Steps must be higher than or equals to 1")
        self._delay = delay
        self._iterator = iter(iterable)
        self._i = 0
        self._steps = steps
        self._map = None

    def __aiter__(self) -> AsyncIter[_T]:
        return self

    async def __anext__(self) -> _T:
        try:
            item = next(self._iterator)
        except StopIteration as e:
            raise StopAsyncIteration from e
        if self._i == self._steps:
            self._i = 0
            await asyncio.sleep(self._delay)
        self._i += 1
        return await discord.utils.maybe_coroutine(self._map, item) if self._map is not None else item

    def __await__(self) -> Generator[Any, None, list[_T]]:
        """Returns a list of the iterable.

        Examples
        --------
        >>> from pylav.utils.vendor.redbot import AsyncIter
        >>> iterator = AsyncIter(range(5))
        >>> await iterator
        [0, 1, 2, 3, 4]

        """
        return self.flatten().__await__()

    async def next(self, default: Any = ...) -> _T:
        """Returns a next entry of the iterable.

        Parameters
        ----------
        default: Optional[Any]
            The value to return if the iterator is exhausted.

        Raises
        ------
        StopAsyncIteration
            When ``default`` is not specified and the iterator has been exhausted.

        Examples
        --------
        >>> from pylav.utils.vendor.redbot import AsyncIter
        >>> iterator = AsyncIter(range(5))
        >>> await iterator.next()
        0
        >>> await iterator.next()
        1

        """
        try:
            value = await self.__anext__()
        except StopAsyncIteration:
            if default is ...:
                raise
            value = default
        return value

    async def flatten(self) -> list[_T]:
        """Returns a list of the iterable.

        Examples
        --------
        >>> from pylav.utils.vendor.redbot import AsyncIter
        >>> iterator = AsyncIter(range(5))
        >>> await iterator.flatten()
        [0, 1, 2, 3, 4]

        """
        return [item async for item in self]

    def filter(self, function: Callable[[_T], bool | Awaitable[bool]]) -> AsyncFilter[_T]:
        """Filter the iterable with an (optionally async) predicate.

        Parameters
        ----------
        function: Callable[[T], Union[bool, Awaitable[bool]]]
            A function or coroutine function which takes one item of ``iterable``
            as an argument, and returns ``True`` or ``False``.

        Returns
        -------
        AsyncFilter[T]
            An object which can either be awaited to yield a list of the filtered
            items, or can also act as an async iterator to yield items one by one.

        Examples
        --------
        >>> from pylav.utils.vendor.redbot import AsyncIter
        >>> def predicate(value):
        ...     return value <= 5
        >>> iterator = AsyncIter([1, 10, 5, 100])
        >>> async for i in iterator.filter(predicate):
        ...     print(i)
        1
        5

        >>> from pylav.utils.vendor.redbot import AsyncIter
        >>> def predicate(value):
        ...     return value <= 5
        >>> iterator = AsyncIter([1, 10, 5, 100])
        >>> await iterator.filter(predicate)
        [1, 5]

        """
        return async_filter(function, self)

    def enumerate(self, start: int = 0) -> AsyncIterator[tuple[int, _T]]:
        """Async iterable version of `enumerate`.

        Parameters
        ----------
        start: int
            The index to start from. Defaults to 0.

        Returns
        -------
        AsyncIterator[Tuple[int, T]]
            An async iterator of tuples in the form of ``(index, item)``.

        Examples
        --------
        >>> from pylav.utils.vendor.redbot import AsyncIter
        >>> iterator = AsyncIter(['one', 'two', 'three'])
        >>> async for i in iterator.enumerate(start=10):
        ...     print(i)
        (10, 'one')
        (11, 'two')
        (12, 'three')

        """
        return async_enumerate(self, start)

    async def without_duplicates(self) -> AsyncIterator[_T]:
        """
        Iterates while omitting duplicated entries.

        Examples
        --------
        >>> from pylav.utils.vendor.redbot import AsyncIter
        >>> iterator = AsyncIter([1,2,3,3,4,4,5])
        >>> async for i in iterator.without_duplicates():
        ...     print(i)
        1
        2
        3
        4
        5

        """
        _temp = set()
        async for item in self:
            if item not in _temp:
                yield item
                _temp.add(item)
        del _temp

    async def find(
        self,
        predicate: Callable[[_T], bool | Awaitable[bool]],
        default: Any | None = None,
    ) -> AsyncIterator[_T]:
        """Calls ``predicate`` over items in iterable and return first value to match.

        Parameters
        ----------
        predicate: Union[Callable, Coroutine]
            A function that returns a boolean-like result. The predicate provided can be a coroutine.
        default: Optional[Any]
            The value to return if there are no matches.

        Raises
        ------
        TypeError
            When ``predicate`` is not a callable.

        Examples
        --------
        >>> from pylav.utils.vendor.redbot import AsyncIter
        >>> await AsyncIter(range(3)).find(lambda x: x == 1)
        1
        """
        while True:
            try:
                elem = await self.__anext__()
            except StopAsyncIteration:
                return default
            ret = await discord.utils.maybe_coroutine(predicate, elem)
            if ret:
                return elem

    def map(self, func: Callable[[_T], _S | Awaitable[_S]]) -> AsyncIter[_S]:
        """Set the mapping callable for this instance of `AsyncIter`.

        .. important::
            This should be called after AsyncIter initialization and before any other of its methods.

        Parameters
        ----------
        func: Union[Callable, Coroutine]
            The function to map values to. The function provided can be a coroutine.

        Raises
        ------
        TypeError
            When ``func`` is not a callable.

        Examples
        --------
        >>> from pylav.utils.vendor.redbot import AsyncIter
        >>> async for value in AsyncIter(range(3)).map(bool):
        ...     print(value)
        False
        True
        True

        """

        if not callable(func):
            raise TypeError("Mapping must be a callable")
        self._map = func
        return self


class MessagePredicate(Callable[[discord.Message], bool]):
    """A simple collection of predicates for message events.

    These predicates intend to help simplify checks in message events
    and reduce boilerplate code.

    This class should be created through the provided classmethods.
    Instances of this class are callable message predicates, i.e. they
    return ``True`` if a message matches the criteria.

    All predicates are combined with :meth:`MessagePredicate.same_context`.

    Examples
    --------
    Waiting for a response in the same channel and from the same
    author::

        await bot.wait_for("message", check=MessagePredicate.same_context(ctx))

    Waiting for a response to a yes or no question::

        pred = MessagePredicate.yes_or_no(ctx)
        await bot.wait_for("message", check=pred)
        if pred.result is True:
            # User responded "yes"
            ...

    Getting a member object from a user's response::

        pred = MessagePredicate.valid_member(ctx)
        await bot.wait_for("message", check=pred)
        member = pred.result

    Attributes
    ----------
    result : Any
        The object which the message content matched with. This is
        dependent on the predicate used - see each predicate's
        documentation for details, not every method will assign this
        attribute. Defaults to ``None``.

    """

    __slots__ = ("result", "_pred")

    def __init__(self, predicate: Callable[[MessagePredicate, discord.Message], bool]) -> None:
        self._pred: Callable[[MessagePredicate, discord.Message], bool] = predicate
        self.result = None

    def __call__(self, message: discord.Message) -> bool:
        return self._pred(self, message)

    @classmethod
    def same_context(
        cls,
        ctx: commands.Context | None = None,
        channel: discord.TextChannel | discord.Thread | discord.DMChannel | None = None,
        user: discord.abc.User | None = None,
    ) -> MessagePredicate:
        """Match if the message fits the described context.

        Parameters
        ----------
        ctx : Optional[Context]
            The current invocation context.
        channel : Optional[Union[`discord.TextChannel`, `discord.Thread`, `discord.DMChannel`]]
            The channel we expect a message in. If unspecified,
            defaults to ``ctx.channel``. If ``ctx`` is unspecified
            too, the message's channel will be ignored.
        user : Optional[discord.abc.User]
            The user we expect a message from. If unspecified,
            defaults to ``ctx.author``. If ``ctx`` is unspecified
            too, the message's author will be ignored.

        Returns
        -------
        MessagePredicate
            The event predicate.

        """
        if ctx is not None:
            channel = channel or ctx.channel
            user = user or ctx.author

        return cls(
            lambda self, m: (user is None or user.id == m.author.id) and (channel is None or channel.id == m.channel.id)
        )

    @classmethod
    def cancelled(
        cls,
        ctx: commands.Context | None = None,
        channel: discord.TextChannel | discord.Thread | discord.DMChannel | None = None,
        user: discord.abc.User | None = None,
    ) -> MessagePredicate:
        """Match if the message is ``[p]cancel``.

        Parameters
        ----------
        ctx : Optional[Context]
            Same as ``ctx`` in :meth:`same_context`.
        channel : Optional[Union[`discord.TextChannel`, `discord.Thread`, `discord.DMChannel`]]
            Same as ``channel`` in :meth:`same_context`.
        user : Optional[discord.abc.User]
            Same as ``user`` in :meth:`same_context`.

        Returns
        -------
        MessagePredicate
            The event predicate.

        """
        same_context = cls.same_context(ctx, channel, user)
        return cls(lambda self, m: (same_context(m) and m.content.lower() == f"{ctx.prefix}cancel"))

    @classmethod
    def yes_or_no(
        cls,
        ctx: commands.Context | None = None,
        channel: discord.TextChannel | discord.Thread | discord.DMChannel | None = None,
        user: discord.abc.User | None = None,
    ) -> MessagePredicate:
        """Match if the message is "yes"/"y" or "no"/"n".

        This will assign ``True`` for *yes*, or ``False`` for *no* to
        the `result` attribute.

        Parameters
        ----------
        ctx : Optional[Context]
            Same as ``ctx`` in :meth:`same_context`.
        channel : Optional[Union[`discord.TextChannel`, `discord.Thread`, `discord.DMChannel`]]
            Same as ``channel`` in :meth:`same_context`.
        user : Optional[discord.abc.User]
            Same as ``user`` in :meth:`same_context`.

        Returns
        -------
        MessagePredicate
            The event predicate.

        """
        same_context = cls.same_context(ctx, channel, user)

        def predicate(self: MessagePredicate, m: discord.Message) -> bool:
            if not same_context(m):
                return False
            content = m.content.lower()
            if content in ("yes", "y"):
                self.result = True
            elif content in ("no", "n"):
                self.result = False
            else:
                return False
            return True

        return cls(predicate)

    @classmethod
    def valid_int(
        cls,
        ctx: commands.Context | None = None,
        channel: discord.TextChannel | discord.Thread | discord.DMChannel | None = None,
        user: discord.abc.User | None = None,
    ) -> MessagePredicate:
        """Match if the response is an integer.

        Assigns the response to `result` as an `int`.

        Parameters
        ----------
        ctx : Optional[Context]
            Same as ``ctx`` in :meth:`same_context`.
        channel : Optional[Union[`discord.TextChannel`, `discord.Thread`, `discord.DMChannel`]]
            Same as ``channel`` in :meth:`same_context`.
        user : Optional[discord.abc.User]
            Same as ``user`` in :meth:`same_context`.

        Returns
        -------
        MessagePredicate
            The event predicate.

        """
        same_context = cls.same_context(ctx, channel, user)

        def predicate(self: MessagePredicate, m: discord.Message) -> bool:
            if not same_context(m):
                return False
            try:
                self.result = int(m.content)
            except ValueError:
                return False
            else:
                return True

        return cls(predicate)

    @classmethod
    def valid_float(
        cls,
        ctx: commands.Context | None = None,
        channel: discord.TextChannel | discord.Thread | discord.DMChannel | None = None,
        user: discord.abc.User | None = None,
    ) -> MessagePredicate:
        """Match if the response is a float.

        Assigns the response to `result` as a `float`.

        Parameters
        ----------
        ctx : Optional[Context]
            Same as ``ctx`` in :meth:`same_context`.
        channel : Optional[Union[`discord.TextChannel`, `discord.Thread`, `discord.DMChannel`]]
            Same as ``channel`` in :meth:`same_context`.
        user : Optional[discord.abc.User]
            Same as ``user`` in :meth:`same_context`.

        Returns
        -------
        MessagePredicate
            The event predicate.

        """
        same_context = cls.same_context(ctx, channel, user)

        def predicate(self: MessagePredicate, m: discord.Message) -> bool:
            if not same_context(m):
                return False
            try:
                self.result = float(m.content)
            except ValueError:
                return False
            else:
                return True

        return cls(predicate)

    @classmethod
    def positive(
        cls,
        ctx: commands.Context | None = None,
        channel: discord.TextChannel | discord.Thread | discord.DMChannel | None = None,
        user: discord.abc.User | None = None,
    ) -> MessagePredicate:
        """Match if the response is a positive number.

        Assigns the response to `result` as a `float`.

        Parameters
        ----------
        ctx : Optional[Context]
            Same as ``ctx`` in :meth:`same_context`.
        channel : Optional[Union[`discord.TextChannel`, `discord.Thread`, `discord.DMChannel`]]
            Same as ``channel`` in :meth:`same_context`.
        user : Optional[discord.abc.User]
            Same as ``user`` in :meth:`same_context`.

        Returns
        -------
        MessagePredicate
            The event predicate.

        """
        same_context = cls.same_context(ctx, channel, user)

        def predicate(self: MessagePredicate, m: discord.Message) -> bool:
            if not same_context(m):
                return False
            try:
                number = float(m.content)
            except ValueError:
                return False
            else:
                if number > 0:
                    self.result = number
                    return True
                else:
                    return False

        return cls(predicate)

    @classmethod
    def valid_role(
        cls,
        ctx: commands.Context | None = None,
        channel: discord.TextChannel | discord.Thread | None = None,
        user: discord.abc.User | None = None,
    ) -> MessagePredicate:
        """Match if the response refers to a role in the current guild.

        Assigns the matching `discord.Role` object to `result`.

        This predicate cannot be used in DM.

        Parameters
        ----------
        ctx : Optional[Context]
            Same as ``ctx`` in :meth:`same_context`.
        channel : Optional[Union[`discord.TextChannel`, `discord.Thread`]]
            Same as ``channel`` in :meth:`same_context`.
        user : Optional[discord.abc.User]
            Same as ``user`` in :meth:`same_context`.

        Returns
        -------
        MessagePredicate
            The event predicate.

        """
        same_context = cls.same_context(ctx, channel, user)
        guild = cls._get_guild(ctx, channel, cast(discord.Member, user))

        def predicate(self: MessagePredicate, m: discord.Message) -> bool:
            if not same_context(m):
                return False

            role = self._find_role(guild, m.content)
            if role is None:
                return False

            self.result = role
            return True

        return cls(predicate)

    @classmethod
    def valid_member(
        cls,
        ctx: commands.Context | None = None,
        channel: discord.TextChannel | discord.Thread | None = None,
        user: discord.abc.User | None = None,
    ) -> MessagePredicate:
        """Match if the response refers to a member in the current guild.

        Assigns the matching `discord.Member` object to `result`.

        This predicate cannot be used in DM.

        Parameters
        ----------
        ctx : Optional[Context]
            Same as ``ctx`` in :meth:`same_context`.
        channel : Optional[Union[`discord.TextChannel`, `discord.Thread`]]
            Same as ``channel`` in :meth:`same_context`.
        user : Optional[discord.abc.User]
            Same as ``user`` in :meth:`same_context`.

        Returns
        -------
        MessagePredicate
            The event predicate.

        """
        same_context = cls.same_context(ctx, channel, user)
        guild = cls._get_guild(ctx, channel, cast(discord.Member, user))

        def predicate(self: MessagePredicate, m: discord.Message) -> bool:
            if not same_context(m):
                return False

            match = DISCORD_ID.match(m.content) or DISCORD_USER_MENTION.match(m.content)
            if match:
                result = guild.get_member(int(match.group(1)))
            else:
                result = guild.get_member_named(m.content)

            if result is None:
                return False
            self.result = result
            return True

        return cls(predicate)

    @classmethod
    def valid_text_channel(
        cls,
        ctx: commands.Context | None = None,
        channel: discord.TextChannel | discord.Thread | None = None,
        user: discord.abc.User | None = None,
    ) -> MessagePredicate:
        """Match if the response refers to a text channel in the current guild.

        Assigns the matching `discord.TextChannel` object to `result`.

        This predicate cannot be used in DM.

        Parameters
        ----------
        ctx : Optional[Context]
            Same as ``ctx`` in :meth:`same_context`.
        channel : Optional[Union[`discord.TextChannel`, `discord.Thread`]]
            Same as ``channel`` in :meth:`same_context`.
        user : Optional[discord.abc.User]
            Same as ``user`` in :meth:`same_context`.

        Returns
        -------
        MessagePredicate
            The event predicate.

        """
        same_context = cls.same_context(ctx, channel, user)
        guild = cls._get_guild(ctx, channel, cast(discord.Member, user))

        def predicate(self: MessagePredicate, m: discord.Message) -> bool:
            if not same_context(m):
                return False

            match = DISCORD_ID.match(m.content) or DISCORD_CHANNEL_MENTION.match(m.content)
            if match:
                result = guild.get_channel(int(match.group(1)))
            else:
                result = discord.utils.get(guild.text_channels, name=m.content)

            if not isinstance(result, discord.TextChannel):
                return False
            self.result = result
            return True

        return cls(predicate)

    @classmethod
    def has_role(
        cls,
        ctx: commands.Context | None = None,
        channel: discord.TextChannel | discord.Thread | None = None,
        user: discord.abc.User | None = None,
    ) -> MessagePredicate:
        """Match if the response refers to a role which the author has.

        Assigns the matching `discord.Role` object to `result`.

        One of ``user`` or ``ctx`` must be supplied. This predicate
        cannot be used in DM.

        Parameters
        ----------
        ctx : Optional[Context]
            Same as ``ctx`` in :meth:`same_context`.
        channel : Optional[Union[`discord.TextChannel`, `discord.Thread`]]
            Same as ``channel`` in :meth:`same_context`.
        user : Optional[discord.abc.User]
            Same as ``user`` in :meth:`same_context`.

        Returns
        -------
        MessagePredicate
            The event predicate.

        """
        same_context = cls.same_context(ctx, channel, user)
        guild = cls._get_guild(ctx, channel, cast(discord.Member, user))
        if user is None:
            if ctx is None:
                raise TypeError("One of `user` or `ctx` must be supplied to `MessagePredicate.has_role`.")
            user = ctx.author

        def predicate(self: MessagePredicate, m: discord.Message) -> bool:
            if not same_context(m):
                return False

            role = self._find_role(guild, m.content)
            if role is None or role not in user.roles:
                return False

            self.result = role
            return True

        return cls(predicate)

    @classmethod
    def equal_to(
        cls,
        value: str,
        ctx: commands.Context | None = None,
        channel: discord.TextChannel | discord.Thread | discord.DMChannel | None = None,
        user: discord.abc.User | None = None,
    ) -> MessagePredicate:
        """Match if the response is equal to the specified value.

        Parameters
        ----------
        value : str
            The value to compare the response with.
        ctx : Optional[Context]
            Same as ``ctx`` in :meth:`same_context`.
        channel : Optional[Union[`discord.TextChannel`, `discord.Thread`, `discord.DMChannel`]]
            Same as ``channel`` in :meth:`same_context`.
        user : Optional[discord.abc.User]
            Same as ``user`` in :meth:`same_context`.

        Returns
        -------
        MessagePredicate
            The event predicate.

        """
        same_context = cls.same_context(ctx, channel, user)
        return cls(lambda self, m: same_context(m) and m.content == value)

    @classmethod
    def lower_equal_to(
        cls,
        value: str,
        ctx: commands.Context | None = None,
        channel: discord.TextChannel | discord.Thread | discord.DMChannel | None = None,
        user: discord.abc.User | None = None,
    ) -> MessagePredicate:
        """Match if the response *as lowercase* is equal to the specified value.

        Parameters
        ----------
        value : str
            The value to compare the response with.
        ctx : Optional[Context]
            Same as ``ctx`` in :meth:`same_context`.
        channel : Optional[Union[`discord.TextChannel`, `discord.Thread`, `discord.DMChannel`]]
            Same as ``channel`` in :meth:`same_context`.
        user : Optional[discord.abc.User]
            Same as ``user`` in :meth:`same_context`.

        Returns
        -------
        MessagePredicate
            The event predicate.

        """
        same_context = cls.same_context(ctx, channel, user)
        return cls(lambda self, m: same_context(m) and m.content.lower() == value)

    @classmethod
    def less(
        cls,
        value: int | float,
        ctx: commands.Context | None = None,
        channel: discord.TextChannel | discord.Thread | discord.DMChannel | None = None,
        user: discord.abc.User | None = None,
    ) -> MessagePredicate:
        """Match if the response is less than the specified value.

        Parameters
        ----------
        value : Union[int, float]
            The value to compare the response with.
        ctx : Optional[Context]
            Same as ``ctx`` in :meth:`same_context`.
        channel : Optional[Union[`discord.TextChannel`, `discord.Thread`, `discord.DMChannel`]]
            Same as ``channel`` in :meth:`same_context`.
        user : Optional[discord.abc.User]
            Same as ``user`` in :meth:`same_context`.

        Returns
        -------
        MessagePredicate
            The event predicate.

        """
        valid_int = cls.valid_int(ctx, channel, user)
        valid_float = cls.valid_float(ctx, channel, user)
        return cls(lambda self, m: (valid_int(m) or valid_float(m)) and float(m.content) < value)

    @classmethod
    def greater(
        cls,
        value: int | float,
        ctx: commands.Context | None = None,
        channel: discord.TextChannel | discord.Thread | discord.DMChannel | None = None,
        user: discord.abc.User | None = None,
    ) -> MessagePredicate:
        """Match if the response is greater than the specified value.

        Parameters
        ----------
        value : Union[int, float]
            The value to compare the response with.
        ctx : Optional[Context]
            Same as ``ctx`` in :meth:`same_context`.
        channel : Optional[Union[`discord.TextChannel`, `discord.Thread`, `discord.DMChannel`]]
            Same as ``channel`` in :meth:`same_context`.
        user : Optional[discord.abc.User]
            Same as ``user`` in :meth:`same_context`.

        Returns
        -------
        MessagePredicate
            The event predicate.

        """
        valid_int = cls.valid_int(ctx, channel, user)
        valid_float = cls.valid_float(ctx, channel, user)
        return cls(lambda self, m: (valid_int(m) or valid_float(m)) and float(m.content) > value)

    @classmethod
    def length_less(
        cls,
        length: int,
        ctx: commands.Context | None = None,
        channel: discord.TextChannel | discord.Thread | discord.DMChannel | None = None,
        user: discord.abc.User | None = None,
    ) -> MessagePredicate:
        """Match if the response's length is less than the specified length.

        Parameters
        ----------
        length : int
            The value to compare the response's length with.
        ctx : Optional[Context]
            Same as ``ctx`` in :meth:`same_context`.
        channel : Optional[Union[`discord.TextChannel`, `discord.Thread`, `discord.DMChannel`]]
            Same as ``channel`` in :meth:`same_context`.
        user : Optional[discord.abc.User]
            Same as ``user`` in :meth:`same_context`.

        Returns
        -------
        MessagePredicate
            The event predicate.

        """
        same_context = cls.same_context(ctx, channel, user)
        return cls(lambda self, m: same_context(m) and len(m.content) <= length)

    @classmethod
    def length_greater(
        cls,
        length: int,
        ctx: commands.Context | None = None,
        channel: discord.TextChannel | discord.Thread | discord.DMChannel | None = None,
        user: discord.abc.User | None = None,
    ) -> MessagePredicate:
        """Match if the response's length is greater than the specified length.

        Parameters
        ----------
        length : int
            The value to compare the response's length with.
        ctx : Optional[Context]
            Same as ``ctx`` in :meth:`same_context`.
        channel : Optional[Union[`discord.TextChannel`, `discord.Thread`, `discord.DMChannel`]]
            Same as ``channel`` in :meth:`same_context`.
        user : Optional[discord.abc.User]
            Same as ``user`` in :meth:`same_context`.

        Returns
        -------
        MessagePredicate
            The event predicate.

        """
        same_context = cls.same_context(ctx, channel, user)
        return cls(lambda self, m: same_context(m) and len(m.content) >= length)

    @classmethod
    def contained_in(
        cls,
        collection: Sequence[str],
        ctx: commands.Context | None = None,
        channel: discord.TextChannel | discord.Thread | discord.DMChannel | None = None,
        user: discord.abc.User | None = None,
    ) -> MessagePredicate:
        """Match if the response is contained in the specified collection.

        The index of the response in the ``collection`` sequence is
        assigned to the `result` attribute.

        Parameters
        ----------
        collection : Sequence[str]
            The collection containing valid responses.
        ctx : Optional[Context]
            Same as ``ctx`` in :meth:`same_context`.
        channel : Optional[Union[`discord.TextChannel`, `discord.Thread`, `discord.DMChannel`]]
            Same as ``channel`` in :meth:`same_context`.
        user : Optional[discord.abc.User]
            Same as ``user`` in :meth:`same_context`.

        Returns
        -------
        MessagePredicate
            The event predicate.

        """
        same_context = cls.same_context(ctx, channel, user)

        def predicate(self: MessagePredicate, m: discord.Message) -> bool:
            if not same_context(m):
                return False
            try:
                self.result = collection.index(m.content)
            except ValueError:
                return False
            else:
                return True

        return cls(predicate)

    @classmethod
    def lower_contained_in(
        cls,
        collection: Sequence[str],
        ctx: commands.Context | None = None,
        channel: discord.TextChannel | discord.Thread | discord.DMChannel | None = None,
        user: discord.abc.User | None = None,
    ) -> MessagePredicate:
        """Same as :meth:`contained_in`, but the response is set to lowercase before matching.

        Parameters
        ----------
        collection : Sequence[str]
            The collection containing valid lowercase responses.
        ctx : Optional[Context]
            Same as ``ctx`` in :meth:`same_context`.
        channel : Optional[Union[`discord.TextChannel`, `discord.Thread`, `discord.DMChannel`]]
            Same as ``channel`` in :meth:`same_context`.
        user : Optional[discord.abc.User]
            Same as ``user`` in :meth:`same_context`.

        Returns
        -------
        MessagePredicate
            The event predicate.

        """
        same_context = cls.same_context(ctx, channel, user)

        def predicate(self: MessagePredicate, m: discord.Message) -> bool:
            if not same_context(m):
                return False
            try:
                self.result = collection.index(m.content.lower())
            except ValueError:
                return False
            else:
                return True

        return cls(predicate)

    @classmethod
    def regex(
        cls,
        pattern: Pattern[str] | str,
        ctx: commands.Context | None = None,
        channel: discord.TextChannel | discord.Thread | discord.DMChannel | None = None,
        user: discord.abc.User | None = None,
    ) -> MessagePredicate:
        """Match if the response matches the specified regex pattern.

        This predicate will use `re.search` to find a match. The
        resulting `match object <match-objects>` will be assigned
        to `result`.

        Parameters
        ----------
        pattern : Union[`pattern object <re-objects>`, str]
            The pattern to search for in the response.
        ctx : Optional[Context]
            Same as ``ctx`` in :meth:`same_context`.
        channel : Optional[Union[`discord.TextChannel`, `discord.Thread`, `discord.DMChannel`]]
            Same as ``channel`` in :meth:`same_context`.
        user : Optional[discord.abc.User]
            Same as ``user`` in :meth:`same_context`.

        Returns
        -------
        MessagePredicate
            The event predicate.

        """
        same_context = cls.same_context(ctx, channel, user)

        def predicate(self: MessagePredicate, m: discord.Message) -> bool:
            if not same_context(m):
                return False
            pattern_obj = re.compile(pattern) if isinstance(pattern, str) else pattern
            if match := pattern_obj.search(m.content):
                self.result = match
                return True
            return False

        return cls(predicate)

    @staticmethod
    def _find_role(guild: discord.Guild, argument: str) -> discord.Role | None:
        return (
            guild.get_role(int(match.group(1)))
            if (match := DISCORD_ID.match(argument) or DISCORD_ROLE_MENTION.match(argument))
            else discord.utils.get(guild.roles, name=argument)
        )

    @staticmethod
    def _get_guild(
        ctx: commands.Context | None,
        channel: discord.TextChannel | discord.Thread | None,
        user: discord.Member | None,
    ) -> discord.Guild:
        if ctx is not None:
            return ctx.guild
        elif channel is not None:
            return channel.guild
        elif user is not None:
            return user.guild



from __future__ import annotations

from dataclasses import dataclass

from piccolo.querystring import QueryString

from pylav.compat import json
from pylav.constants.builtin_nodes import BUNDLED_NODES_IDS_HOST_MAPPING
from pylav.constants.config import JAVA_EXECUTABLE
from pylav.constants.node import NODE_DEFAULT_SETTINGS
from pylav.constants.node_features import SUPPORTED_FEATURES, SUPPORTED_SOURCES
from pylav.extension.bundled_node.utils import get_jar_ram_actual
from pylav.helpers.singleton import SingletonCachedByKey
from pylav.storage.database.cache.decodators import maybe_cached
from pylav.storage.database.cache.model import CachedModel
from pylav.storage.database.tables.nodes import NodeRow, Sessions
from pylav.type_hints.dict_typing import JSON_DICT_TYPE


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True, frozen=True)
class Node(CachedModel, metaclass=SingletonCachedByKey):
    id: int

    def get_cache_key(self) -> str:
        return f"{self.id}"

    @maybe_cached
    async def exists(self) -> bool:
        """Check if the node exists in the database.

        Returns
        -------
        bool
            Whether the node exists in the database.
        """
        return await NodeRow.exists().where(NodeRow.id == self.id)

    async def delete(self) -> None:
        """Delete the node from the database"""
        await NodeRow.delete().where(NodeRow.id == self.id)
        await self.invalidate_cache()

    @maybe_cached
    async def fetch_all(self) -> JSON_DICT_TYPE:
        data = await NodeRow.select().where(NodeRow.id == self.id).first().output(load_json=True, nested=True)
        return data or {
            "id": self.id,
            "name": NodeRow.name.default,
            "ssl": NodeRow.ssl.default,
            "resume_timeout": NodeRow.resume_timeout.default,
            "reconnect_attempts": NodeRow.reconnect_attempts.default,
            "search_only": NodeRow.search_only.default,
            "managed": NodeRow.managed.default,
            "extras": json.loads(NodeRow.extras.default),
            "yaml": json.loads(NodeRow.yaml.default),
            "disabled_sources": NodeRow.disabled_sources.default,
        }

    @maybe_cached
    async def fetch_name(self) -> str | None:
        """Fetch the node from the database.

        Returns
        -------
        str
            The node's name.
        """
        data = (
            await NodeRow.select(NodeRow.name).where(NodeRow.id == self.id).first().output(load_json=True, nested=True)
        )
        return data["name"] if data else None

    async def update_name(self, name: str) -> None:
        """Update the node's name in the database"""
        await NodeRow.insert(NodeRow(id=self.id, name=name)).on_conflict(
            action="DO UPDATE", target=NodeRow.id, values=[NodeRow.name]
        )
        await self.update_cache((self.fetch_name, name), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_ssl(self) -> bool:
        """Fetch the node's ssl setting from the database.

        Returns
        -------
        bool
            The node's ssl setting.
        """
        data = (
            await NodeRow.select(NodeRow.ssl).where(NodeRow.id == self.id).first().output(load_json=True, nested=True)
        )
        return data["ssl"] if data else NodeRow.ssl.default

    async def update_ssl(self, ssl: bool) -> None:
        """Update the node's ssl setting in the database"""
        await NodeRow.insert(NodeRow(id=self.id, ssl=ssl)).on_conflict(
            action="DO UPDATE", target=NodeRow.id, values=[NodeRow.ssl]
        )
        await self.update_cache((self.fetch_ssl, ssl), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_resume_timeout(self) -> int:
        """Fetch the node's resume timeout from the database.

        Returns
        -------
        int
            The node's resume timeout.
        """
        data = (
            await NodeRow.select(NodeRow.resume_timeout)
            .where(NodeRow.id == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return data["resume_timeout"] if data else NodeRow.resume_timeout.default

    async def update_resume_timeout(self, resume_timeout: int) -> None:
        """Update the node's resume timeout in the database"""
        await NodeRow.insert(NodeRow(id=self.id, resume_timeout=resume_timeout)).on_conflict(
            action="DO UPDATE", target=NodeRow.id, values=[NodeRow.resume_timeout]
        )
        await self.update_cache((self.fetch_resume_timeout, resume_timeout), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_reconnect_attempts(self) -> int:
        """Fetch the node's reconnect attempts from the database.

        Returns
        -------
        int
            The node's reconnect attempts.
        """
        data = (
            await NodeRow.select(NodeRow.reconnect_attempts)
            .where(NodeRow.id == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return data["reconnect_attempts"] if data else NodeRow.reconnect_attempts.default

    async def update_reconnect_attempts(self, reconnect_attempts: int) -> None:
        """Update the node's reconnect attempts in the database"""
        await NodeRow.insert(NodeRow(id=self.id, reconnect_attempts=reconnect_attempts)).on_conflict(
            action="DO UPDATE", target=NodeRow.id, values=[NodeRow.reconnect_attempts]
        )
        await self.update_cache((self.fetch_reconnect_attempts, reconnect_attempts), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_search_only(self) -> bool:
        """Fetch the node's search only setting from the database.

        Returns
        -------
        bool
            The node's search only setting.
        """
        data = (
            await NodeRow.select(NodeRow.search_only)
            .where(NodeRow.id == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return data["search_only"] if data else NodeRow.search_only.default

    async def update_search_only(self, search_only: bool) -> None:
        """Update the node's search only setting in the database"""
        await NodeRow.insert(NodeRow(id=self.id, search_only=search_only)).on_conflict(
            action="DO UPDATE", target=NodeRow.id, values=[NodeRow.search_only]
        )
        await self.update_cache((self.fetch_search_only, search_only), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_managed(self) -> bool:
        """Fetch the node's managed setting from the database.

        Returns
        -------
        bool
            The node's managed setting.
        """
        data = (
            await NodeRow.select(NodeRow.managed)
            .where(NodeRow.id == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return data["managed"] if data else NodeRow.managed.default

    async def update_managed(self, managed: bool) -> None:
        """Update the node's managed setting in the database"""
        await NodeRow.insert(NodeRow(id=self.id, managed=managed)).on_conflict(
            action="DO UPDATE", target=NodeRow.id, values=[NodeRow.managed]
        )
        await self.update_cache((self.fetch_managed, managed), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_extras(self) -> JSON_DICT_TYPE:
        """Fetch the node's extras from the database.

        Returns
        -------
        dict
            The node's extras.
        """
        data = (
            await NodeRow.select(NodeRow.extras)
            .where(NodeRow.id == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return data["extras"] if data else json.loads(NodeRow.extras.default)

    async def update_extras(self, extras: JSON_DICT_TYPE) -> None:
        """Update the node's extras in the database"""
        await NodeRow.insert(NodeRow(id=self.id, extras=extras)).on_conflict(
            action="DO UPDATE", target=NodeRow.id, values=[NodeRow.extras]
        )
        await self.update_cache((self.fetch_extras, extras), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_yaml(self) -> JSON_DICT_TYPE:
        """Fetch the node's yaml from the database.

        Returns
        -------
        dict
            The node's yaml.
        """
        data = (
            await NodeRow.select(NodeRow.yaml).where(NodeRow.id == self.id).first().output(load_json=True, nested=True)
        )
        return data["yaml"] if data else json.loads(NodeRow.yaml.default)

    async def update_yaml(self, yaml_data: JSON_DICT_TYPE) -> None:
        """Update the node's yaml in the database"""
        await NodeRow.insert(NodeRow(id=self.id, yaml=yaml_data)).on_conflict(
            action="DO UPDATE", target=NodeRow.id, values=[NodeRow.yaml]
        )
        await self.update_cache((self.fetch_yaml, yaml_data), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_disabled_sources(self) -> list[str]:
        """Fetch the node's disabled sources from the database.

        Returns
        -------
        list[str]
            The node's disabled sources.
        """
        data = (
            await NodeRow.select(NodeRow.disabled_sources)
            .where(NodeRow.id == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return data["disabled_sources"] if data else NodeRow.disabled_sources.default

    async def update_disabled_sources(self, disabled_sources: list[str]) -> None:
        """Update the node's disabled sources in the database"""
        source = set(map(str.strip, map(str.lower, disabled_sources)))
        intersection = list(source & SUPPORTED_SOURCES.union(SUPPORTED_FEATURES))
        await NodeRow.insert(NodeRow(id=self.id, disabled_sources=intersection)).on_conflict(
            action="DO UPDATE", target=NodeRow.id, values=[NodeRow.disabled_sources]
        )
        await self.update_cache((self.fetch_disabled_sources, intersection), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    async def add_to_disabled_sources(self, source: str) -> None:
        """Add a source to the node's disabled sources in the database"""
        await NodeRow.insert(NodeRow(id=self.id, disabled_sources=[source])).on_conflict(
            action="DO UPDATE",
            target=NodeRow.id,
            values=[
                (NodeRow.disabled_sources, QueryString("array_cat(node.disabled_sources, EXCLUDED.disabled_sources)"))
            ],
        )
        await self.update_cache((self.exists, True))
        await self.invalidate_cache(self.fetch_all, self.fetch_disabled_sources)

    async def remove_from_disabled_sources(self, source: str) -> None:
        """Remove a source from the node's disabled sources in the database"""
        await NodeRow.update(disabled_sources=QueryString("array_remove(disabled_sources, {})", source)).where(
            NodeRow.id == self.id
        )
        await self.update_cache((self.exists, True))
        await self.invalidate_cache(self.fetch_all, self.fetch_disabled_sources)

    async def bulk_add_to_disabled_sources(self, sources: list[str]) -> None:
        """Add sources to the node's disabled sources in the database"""
        source = set(map(str.strip, map(str.lower, [sources])))
        intersection = list(source & SUPPORTED_SOURCES.union(SUPPORTED_FEATURES))
        await NodeRow.insert(NodeRow(id=self.id, disabled_sources=intersection)).on_conflict(
            action="DO UPDATE",
            target=NodeRow.id,
            values=[(NodeRow.disabled_sources, QueryString("node.disabled_sources || EXCLUDED.disabled_sources"))],
        )
        await self.update_cache((self.exists, True))
        await self.invalidate_cache(self.fetch_all, self.fetch_disabled_sources)

    async def bulk_remove_from_disabled_sources(self, sources: list[str]) -> None:
        """Remove sources from the node's disabled sources in the database"""
        if not sources:
            return
        for source in sources:
            await self.remove_from_disabled_sources(source)

    async def bulk_update(
        self,
        host: str,
        port: int,
        password: str,
        resume_timeout: int = 60,
        name: str | None = None,
        reconnect_attempts: int = -1,
        ssl: bool = False,
        search_only: bool = False,
        managed: bool = False,
        extras: JSON_DICT_TYPE | None = None,
        yaml: JSON_DICT_TYPE | None = None,
        disabled_sources: list[str] | None = None,
    ) -> None:
        """Update the node's data in the database"""
        yaml_data = yaml or {"server": {}, "lavalink": {"server": {}}}
        yaml_data["server"]["address"] = host  # type: ignore
        yaml_data["server"]["port"] = port  # type: ignore
        yaml_data["lavalink"]["server"]["password"] = password
        if disabled_sources is None:
            disabled_sources = []
        if extras is None:
            extras = {}
        await NodeRow.insert(
            NodeRow(
                id=self.id,
                name=name,
                ssl=ssl,
                resume_key=None,
                resume_timeout=resume_timeout,
                reconnect_attempts=reconnect_attempts,
                search_only=search_only,
                managed=managed,
                disabled_sources=disabled_sources,
                extras=extras,
                yaml=yaml,
            )
        ).on_conflict(
            action="DO UPDATE",
            target=NodeRow.id,
            values=[
                NodeRow.name,
                NodeRow.ssl,
                NodeRow.resume_timeout,
                NodeRow.reconnect_attempts,
                NodeRow.search_only,
                NodeRow.managed,
                NodeRow.disabled_sources,
                NodeRow.extras,
                NodeRow.yaml,
            ],
        )
        await self.invalidate_cache()

    async def get_connection_args(self) -> dict[str, int | str | bool | None, list[str]]:
        """Get the connection args for the node.

        Returns
        -------
        dict
            The connection args.
        """
        data = await self.fetch_all()

        if self.id in BUNDLED_NODES_IDS_HOST_MAPPING:
            data["yaml"]["lavalink"]["server"]["password"] = self.client._user_id

        return {
            "unique_identifier": self.id,
            "host": data["yaml"]["server"]["address"],
            "port": data["yaml"]["server"]["port"],
            "password": data["yaml"]["lavalink"]["server"]["password"],
            "name": data["name"],
            "ssl": data["ssl"],
            "reconnect_attempts": data["reconnect_attempts"],
            "search_only": data["search_only"],
            "resume_timeout": data["resume_timeout"],
            "disabled_sources": data["disabled_sources"],
            "managed": data["managed"],
        }

    @classmethod
    async def create_managed(cls, identifier: int) -> None:
        """Create the player in the database"""

        __, java_xmx_default, __, __ = get_jar_ram_actual(JAVA_EXECUTABLE)
        await NodeRow.insert(
            NodeRow(
                id=identifier,
                managed=True,
                ssl=False,
                reconnect_attempts=-1,
                search_only=False,
                yaml=NODE_DEFAULT_SETTINGS,
                name="PyLavManagedNode",
                resume_key=None,
                resume_timeout=600,
                extras={"max_ram": java_xmx_default},
            )
        ).on_conflict(action="DO NOTHING")

    @maybe_cached
    async def fetch_session(self) -> str | None:
        """Fetch the node's session from the database"""
        data = (
            await Sessions.select(Sessions.id)
            .where((Sessions.node == self.id) & (Sessions.bot == self.client.bot.user.id))  # noqa: E712
            .first()
            .output(load_json=True, nested=True)
        )
        return data["id"] if data else None

    async def update_session(self, session: str | None) -> None:
        """Update the node's session in the database"""
        await Sessions.insert(Sessions(id=session, node=self.id, bot=self.client.bot.user.id)).on_conflict(
            action="DO UPDATE",
            values=[Sessions.id],
            target=(Sessions.node, Sessions.bot),
        )
        await self.update_cache((self.fetch_session, session))



from __future__ import annotations

from pathlib import Path

import discord
from redbot.core.i18n import Translator

from pylav.helpers import emojis
from pylav.type_hints.bot import DISCORD_COG_TYPE, DISCORD_INTERACTION_TYPE

_ = Translator("PyLav", Path(__file__))


class EqualizerButton(discord.ui.Button):
    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None) -> None:
        super().__init__(
            style=style,
            emoji=emojis.EQUALIZER,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE) -> None:
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        kwargs = await self.view.get_page(self.view.current_page)
        await self.view.prepare()
        attachments = []
        if "file" in kwargs:
            attachments = [kwargs.pop("file")]
        elif "files" in kwargs:
            attachments = kwargs.pop("files")
        if attachments:
            kwargs["attachments"] = attachments
        await self.view.message.edit(view=self.view, **kwargs)



from __future__ import annotations

from typing import Literal

from pylav.exceptions.base import PyLavException
from pylav.nodes.api.responses.errors import LavalinkError


class HTTPException(PyLavException):
    """Base exception for HTTP request errors"""

    loadType: Literal["apiError"] = "apiError"
    data: None = None

    def __init__(self, response: LavalinkError):
        self.response = response

    def __bool__(self):
        return False


class UnauthorizedException(HTTPException):
    """Raised when a REST request fails due to an incorrect password"""

    def __bool__(self):
        return False



from __future__ import annotations

from typing import TYPE_CHECKING

from pylav.constants.config.overrides import (
    LOCAL_TRACKS_FOLDER,
    MANAGED_NODE_APPLE_MUSIC_API_KEY,
    MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE,
    MANAGED_NODE_DEEZER_KEY,
    MANAGED_NODE_SPOTIFY_CLIENT_ID,
    MANAGED_NODE_SPOTIFY_CLIENT_SECRET,
    MANAGED_NODE_SPOTIFY_COUNTRY_CODE,
    MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN,
    USE_BUNDLED_EXTERNAL_PYLAV_NODE,
)
from pylav.storage.migrations.logging import LOGGER

if TYPE_CHECKING:
    from pylav.core.client import Client


async def process_envvar_variables(client: Client) -> None:
    """Process envvar variables."""
    LOGGER.info("Running migration - Envvar variables")

    config = client.node_db_manager.bundled_node_config()
    yaml_data = await config.fetch_yaml()
    updated = False
    if MANAGED_NODE_SPOTIFY_CLIENT_ID is not None:
        yaml_data["plugins"]["lavasrc"]["spotify"]["clientId"] = MANAGED_NODE_SPOTIFY_CLIENT_ID
        updated = True
    if MANAGED_NODE_SPOTIFY_CLIENT_SECRET is not None:
        yaml_data["plugins"]["lavasrc"]["spotify"]["clientSecret"] = MANAGED_NODE_SPOTIFY_CLIENT_SECRET
        updated = True
    if MANAGED_NODE_SPOTIFY_COUNTRY_CODE is not None:
        yaml_data["plugins"]["lavasrc"]["spotify"]["countryCode"] = MANAGED_NODE_SPOTIFY_COUNTRY_CODE
        updated = True
    if MANAGED_NODE_APPLE_MUSIC_API_KEY is not None:
        yaml_data["plugins"]["lavasrc"]["applemusic"]["mediaAPIToken"] = MANAGED_NODE_APPLE_MUSIC_API_KEY
        updated = True
    if MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE is not None:
        yaml_data["plugins"]["lavasrc"]["applemusic"]["countryCode"] = MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE
        updated = True
    if MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN is not None:
        yaml_data["plugins"]["lavasrc"]["yandexmusic"]["accessToken"] = MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN
        updated = True
    if MANAGED_NODE_DEEZER_KEY is not None:
        yaml_data["plugins"]["lavasrc"]["deezer"]["key"] = MANAGED_NODE_DEEZER_KEY
        updated = True
    if updated:
        await config.update_yaml(yaml_data)

    if USE_BUNDLED_EXTERNAL_PYLAV_NODE is not None:
        await client.lib_db_manager.get_config().update_use_bundled_pylav_external(USE_BUNDLED_EXTERNAL_PYLAV_NODE)

    if LOCAL_TRACKS_FOLDER is not None:
        await client.lib_db_manager.get_config().update_localtrack_folder(LOCAL_TRACKS_FOLDER)



from __future__ import annotations

import os
import re
from typing import TYPE_CHECKING

from pylav.constants.config import JAVA_EXECUTABLE
from pylav.extension.bundled_node.utils import get_jar_ram_actual
from pylav.storage.migrations.logging import LOGGER

if TYPE_CHECKING:
    from pylav.core.client import Client


async def set_correct_ram_cap(client: Client) -> None:
    """Set the correct RAM cap for the managed node."""
    LOGGER.debug("Running Managed node RAM cap limiter")
    config = client.node_db_manager.bundled_node_config()
    extras = await config.fetch_extras()
    current_max_ram = extras.get("max_ram")
    if current_max_ram is None:
        return
    # noinspection PyProtectedMember
    config_data = client._lib_config_manager.get_config()
    java_executable = await config_data.fetch_java_path()
    if java_executable != JAVA_EXECUTABLE and os.path.exists(JAVA_EXECUTABLE):
        await config_data.update_java_path(JAVA_EXECUTABLE)
        java_executable = JAVA_EXECUTABLE
    min_ram, max_ram, min_ram_int, max_ram_int = get_jar_ram_actual(java_executable)
    size_name = ("", "K", "M", "G", "T")
    match = re.match(r"(\d+)([KMGT])", current_max_ram)
    if match is None:
        return
    current_max_ram_int = int(match[1])
    current_max_ram_size = match[2]
    current_max_ram_int *= 1024 ** size_name.index(current_max_ram_size)
    if current_max_ram_int > max_ram_int:
        LOGGER.debug(f"Updating maximum RAM allocation from {current_max_ram} to {max_ram}")
        extras["max_ram"] = max_ram
        await config.update_extras(extras)
    LOGGER.debug("RAM cap limiter complete")



from __future__ import annotations

import random
from pathlib import Path
from typing import TYPE_CHECKING

import discord
from dacite import from_dict
from redbot.core.i18n import Translator
from redbot.core.utils.chat_formatting import humanize_number
from redbot.vendored.discord.ext import menus

from pylav.extension.red.ui.selectors.options.playlist import PlaylistOption
from pylav.logging import getLogger
from pylav.nodes.api.responses.track import Track as LavalinkTrack
from pylav.players.tracks.obj import Track
from pylav.storage.models.playlist import Playlist
from pylav.type_hints.bot import DISCORD_COG_TYPE
from pylav.type_hints.dict_typing import JSON_DICT_TYPE

if TYPE_CHECKING:
    from pylav.extension.red.ui.menus.generic import PaginatingMenu
    from pylav.extension.red.ui.menus.playlist import PlaylistPickerMenu

LOGGER = getLogger("PyLav.ext.red.ui.sources.playlist")

_ = Translator("PyLav", Path(__file__))
INF = float("inf")
ASCII_ORDER_SORT = "~" * 100


class PlaylistPickerSource(menus.ListPageSource):
    def __init__(self, guild_id: int, cog: DISCORD_COG_TYPE, pages: list[Playlist], message_str: str):
        pages.sort(key=lambda p: p.id)
        super().__init__(entries=pages, per_page=5)
        self.message_str = message_str
        self.per_page = 5
        self.guild_id = guild_id
        self.select_options: list[PlaylistOption] = []
        self.cog = cog
        self.select_mapping: dict[str, Playlist] = {}

    def get_starting_index_and_page_number(self, menu: PlaylistPickerMenu) -> tuple[int, int]:
        page_num = menu.current_page
        start = page_num * self.per_page
        return start, page_num

    async def format_page(self, menu: PlaylistPickerMenu, playlists: list[Playlist]) -> discord.Embed | str:
        idx_start, page_num = self.get_starting_index_and_page_number(menu)
        page = await self.cog.pylav.construct_embed(messageable=menu.ctx, title=self.message_str)

        total_number_of_entries = len(self.entries)
        current_page = humanize_number(page_num + 1)
        total_number_of_pages = humanize_number(self.get_max_pages())

        match total_number_of_entries:
            case 1:
                message = _("Page 1 / 1 | 1 playlist")
            case 0:
                message = _("Page 1 / 1 | 0 playlists")
            case __:
                message = _(
                    "Page {current_page_variable_do_not_translate} / {total_number_of_pages_variable_do_not_translate} | {total_number_of_entries_variable_do_not_translate} playlists"
                ).format(
                    current_page_variable_do_not_translate=current_page,
                    total_number_of_pages_variable_do_not_translate=total_number_of_pages,
                    total_number_of_entries_variable_do_not_translate=humanize_number(total_number_of_entries),
                )

        page.set_footer(text=message)
        return page

    async def get_page(self, page_number):
        if page_number > self.get_max_pages():
            page_number = 0
        base = page_number * self.per_page
        self.select_options.clear()
        self.select_mapping.clear()
        for i, playlist in enumerate(iter(self.entries[base : base + self.per_page]), start=base):  # noqa: E203
            self.select_options.append(await PlaylistOption.from_playlist(playlist=playlist, index=i, bot=self.cog.bot))
            self.select_mapping[f"{playlist.id}"] = playlist
        return self.entries[base : base + self.per_page]  # noqa: E203

    def get_max_pages(self):
        """:class:`int`: The maximum number of pages required to paginate this sequence"""
        return self._max_pages or 1


class TrackMappingSource(menus.ListPageSource):
    def __init__(
        self,
        guild_id: int,
        cog: DISCORD_COG_TYPE,
        playlist: Playlist,
        author: discord.abc.User,
        entries: list[str],
        per_page: int = 10,
    ):
        super().__init__(entries=entries, per_page=per_page)
        self.cog = cog
        self.author = author
        self.guild_id = guild_id
        self.playlist = playlist

    def is_paginating(self) -> bool:
        return True

    def get_starting_index_and_page_number(self, menu: PaginatingMenu) -> tuple[int, int]:
        page_num = menu.current_page
        start = page_num * self.per_page
        return start, page_num

    async def format_page(self, menu: PaginatingMenu, tracks: list[str | JSON_DICT_TYPE]) -> discord.Embed:
        start_index, page_num = self.get_starting_index_and_page_number(menu)
        padding = len(str(start_index + len(tracks)))
        queue_list = ""
        player = self.cog.pylav.get_player(self.guild_id)
        for track_idx, track in enumerate(tracks, start=start_index + 1):
            if isinstance(track, str):
                track = await Track.build_track(
                    node=random.choice(self.cog.pylav.node_manager.nodes),
                    requester=self.author.id,
                    data=track,
                    query=None,
                    player_instance=player,
                )
            else:
                track = await Track.build_track(
                    node=random.choice(self.cog.pylav.node_manager.nodes),
                    requester=self.author.id,
                    data=from_dict(data_class=LavalinkTrack, data=track),
                    query=None,
                    player_instance=player,
                )
            track_description = await track.get_track_display_name(max_length=50, with_url=True)
            diff = padding - len(str(track_idx))
            queue_list += f"`{track_idx}.{' ' * diff}` {track_description}\n"

        page = await self.cog.pylav.construct_embed(
            title="{translation} __{name}__".format(
                name=await self.playlist.fetch_name(), translation=discord.utils.escape_markdown(_("Tracks in"))
            ),
            description=queue_list,
            messageable=menu.ctx,
        )

        total_number_of_entries = len(self.entries)
        current_page = humanize_number(page_num + 1)
        total_number_of_pages = humanize_number(self.get_max_pages())

        match total_number_of_entries:
            case 1:
                message = _("Page 1 / 1 | 1 track")
            case 0:
                message = _("Page 1 / 1 | 0 tracks")
            case __:
                message = _(
                    "Page {current_page_variable_do_not_translate} / {total_number_of_pages_variable_do_not_translate} | {total_number_of_entries_variable_do_not_translate} tracks"
                ).format(
                    current_page_variable_do_not_translate=current_page,
                    total_number_of_pages_variable_do_not_translate=total_number_of_pages,
                    total_number_of_entries_variable_do_not_translate=humanize_number(total_number_of_entries),
                )

        page.set_footer(text=message)
        return page

    def get_max_pages(self):
        """:class:`int`: The maximum number of pages required to paginate this sequence"""
        return self._max_pages or 1


class PlaylistListSource(menus.ListPageSource):
    def __init__(self, cog: DISCORD_COG_TYPE, pages: list[Playlist]):
        pages.sort(key=lambda p: p.id)
        super().__init__(entries=pages, per_page=5)
        self.cog = cog

    def get_starting_index_and_page_number(self, menu: PaginatingMenu) -> tuple[int, int]:
        page_num = menu.current_page
        start = page_num * self.per_page
        return start, page_num

    async def format_page(self, menu: PaginatingMenu, playlists: list[Playlist]) -> discord.Embed | str:
        idx_start, page_num = self.get_starting_index_and_page_number(menu)
        plist = ""
        space = "\N{EN SPACE}"
        for i, playlist in enumerate(playlists, start=idx_start + 1):
            scope_name = await playlist.get_scope_name(self.cog.bot)
            author_name = await playlist.get_author_name(self.cog.bot) or _("Unknown")
            is_same = scope_name == author_name
            playlist_info = ("\n" + space * 4).join(
                (
                    await playlist.get_name_formatted(with_url=True),
                    _("Identifier: {id_variable_do_not_translate}").format(id_variable_do_not_translate=playlist.id),
                    _("Tracks: {num_variable_do_not_translate}").format(
                        num_variable_do_not_translate=await playlist.size()
                    ),
                    _("Author: {name_variable_do_not_translate}").format(name_variable_do_not_translate=author_name),
                    "\n"
                    if is_same
                    else _("Scope: {scope_variable_do_not_translate}\n").format(
                        scope_variable_do_not_translate=scope_name
                    ),
                )
            )

            plist += f"`{i}.` {playlist_info}"

        embed = await self.cog.pylav.construct_embed(
            messageable=menu.ctx,
            title=_("Playlists you can access in this server:"),
            description=plist,
        )

        total_number_of_entries = len(self.entries)
        current_page = humanize_number(page_num + 1)
        total_number_of_pages = humanize_number(self.get_max_pages())

        match total_number_of_entries:
            case 1:
                message = _("Page 1 / 1 | 1 playlist")
            case 0:
                message = _("Page 1 / 1 | 0 playlists")
            case __:
                message = _(
                    "Page {current_page_variable_do_not_translate} / {total_number_of_pages_variable_do_not_translate} | {total_number_of_entries_variable_do_not_translate} playlists"
                ).format(
                    current_page_variable_do_not_translate=current_page,
                    total_number_of_pages_variable_do_not_translate=total_number_of_pages,
                    total_number_of_entries_variable_do_not_translate=humanize_number(total_number_of_entries),
                )

        embed.set_footer(text=message)
        return embed

    def get_max_pages(self):
        """:class:`int`: The maximum number of pages required to paginate this sequence"""
        return self._max_pages or 1



from __future__ import annotations

from pylav.players.filters.misc import FilterMixin


class Timescale(FilterMixin):
    __slots__ = ("_speed", "_pitch", "_rate", "_default")

    def __init__(self, speed: float | None = None, pitch: float | None = None, rate: float | None = None) -> None:
        super().__init__()
        self.speed = speed
        self.pitch = pitch
        self.rate = rate

    def to_dict(self) -> dict[str, float | bool | None]:
        return {
            "speed": self.speed,
            "pitch": self.pitch,
            "rate": self.rate,
        }

    @classmethod
    def from_dict(cls, data: dict[str, float | bool | None]) -> Timescale:
        return cls(speed=data["speed"], pitch=data["pitch"], rate=data["rate"])

    def __repr__(self) -> str:
        return f"<Timescale: speed={self.speed}, pitch={self.pitch}, rate={self.rate}>"

    @property
    def speed(self) -> float | None:
        return self._speed

    @speed.setter
    def speed(self, v: float | None) -> None:
        self._speed = v

    @property
    def pitch(self) -> float | None:
        return self._pitch

    @pitch.setter
    def pitch(self, v: float | None) -> None:
        self._pitch = v

    @property
    def rate(self) -> float | None:
        return self._rate

    @rate.setter
    def rate(self, v: float | None) -> None:
        self._rate = v

    @classmethod
    def default(cls) -> Timescale:
        return cls()

    def get(self) -> dict[str, float]:
        if self.off:
            return {}
        response = {}
        if self.speed is not None:
            response["speed"] = self.speed
        if self.pitch is not None:
            response["pitch"] = self.pitch
        if self.rate is not None:
            response["rate"] = self.rate
        return response

    def reset(self) -> None:
        self.speed = self.pitch = self.rate = None

    @staticmethod
    def _decrease(value: float | int, percentage: float | int) -> float:
        return value / (1 + percentage)

    @staticmethod
    def _increase(value: float | int, percentage: float | int) -> float:
        return value * (1 + percentage)

    def _get_percentages(self) -> tuple[float, float]:
        if self.speed is None:
            speed = 0
        elif self.speed >= 1:
            speed = self.speed - 1
        else:
            speed = 1 - self.speed
        if self.rate is None:
            rate = 0
        elif self.rate >= 1:
            rate = self.rate - 1
        else:
            rate = 1 - self.rate
        return speed, rate

    def adjust_position(self, position: float | int) -> int:
        if self.speed is None and self.rate is None:
            return int(position)
        speed, rate = self._get_percentages()
        if self.rate is None:
            return int(self._decrease(position, speed) if self.speed >= 1 else self._increase(position, speed))
        if self.speed is None:
            return int(
                self._decrease(position, self.rate - 1) if self.rate >= 1 else self._increase(position, 1 - self.rate)
            )
        if self.speed >= 1 and self.rate >= 1:
            return int(self._decrease(position, (self.speed - 1) + (self.rate - 1)))
        elif self.speed >= 1 > self.rate:
            return int(self._decrease(self._increase(position, 1 - self.rate), self.speed - 1))

        elif self.speed < 1 <= self.rate:
            return int(self._decrease(self._increase(position, self.speed - 1), 1 - self.rate))
        else:
            return int(self._increase(position, (1 - self.speed) + (1 - self.rate)))

    def reverse_position(self, position: float | int) -> int:
        if self.speed is None and self.rate is None:
            return int(position)
        speed, rate = self._get_percentages()
        if self.rate is None:
            return int(self._increase(position, speed) if self.speed >= 1 else self._decrease(position, speed))
        if self.speed is None:
            return int(
                self._increase(position, self.rate - 1) if self.rate >= 1 else self._decrease(position, 1 - self.rate)
            )
        if self.speed >= 1 and self.rate >= 1:
            return int(self._increase(position, (self.speed - 1) + (self.rate - 1)))
        elif self.speed >= 1 > self.rate:
            return int(self._increase(self._decrease(position, 1 - self.rate), self.speed - 1))

        elif self.speed < 1 <= self.rate:
            return int(self._increase(self._decrease(position, self.speed - 1), 1 - self.rate))
        else:
            return int(self._decrease(position, (1 - self.speed) + (1 - self.rate)))



from __future__ import annotations

from typing import TYPE_CHECKING

import asyncpg

from pylav.storage.database.tables.misc import DATABASE_ENGINE
from pylav.storage.migrations.logging import LOGGER
from pylav.storage.migrations.low_level.v_1_0_0 import (
    low_level_v_1_0_0_migration,
    migrate_lib_config_v_1_0_0,
    migrate_node_config_v_1_0_0,
    migrate_player_config_v_1_0_0,
    migrate_playlists_v_1_0_0,
    migrate_queries_v_1_0_0,
)
from pylav.storage.migrations.low_level.v_1_3_8 import low_level_v_1_3_8_migration
from pylav.storage.migrations.low_level.v_1_7_0 import low_level_v_1_7_0_migration

if TYPE_CHECKING:
    from pylav.storage.controllers.config import ConfigController


async def run_low_level_migrations(migrator: ConfigController) -> dict[str, dict[str, list[asyncpg.Record] | None]]:
    """
    Runs migrations.
    """
    migration_data = {}
    con = await DATABASE_ENGINE.get_new_connection()
    await low_level_v_1_0_0_migration(con, migration_data, migrator)
    await low_level_v_1_3_8_migration(con)
    await low_level_v_1_7_0_migration(con)
    return migration_data


async def migrate_data(data: dict[str, dict[str, list[asyncpg.Record]]]) -> None:
    """
    Migrates data.
    """

    for version, migrations in data.items():
        match version:
            case "1.0.0":
                if "playlist" in migrations and migrations["playlist"]:
                    LOGGER.info("----------- Migrating Playlist data to PyLav 1.0.0 ---------")
                    await migrate_playlists_v_1_0_0(migrations["playlist"])
                if "query" in migrations and migrations["query"]:
                    LOGGER.info("----------- Migrating Query data to PyLav 1.0.0 ---------")
                    await migrate_queries_v_1_0_0(migrations["query"])
                if "player" in migrations and migrations["player"]:
                    LOGGER.info("----------- Migrating Player settings to PyLav 1.0.0 ---------")
                    await migrate_player_config_v_1_0_0(migrations["player"])
                if "lib" in migrations and migrations["lib"]:
                    LOGGER.info("----------- Migrating Lib settings to PyLav 1.0.0 ---------")
                    await migrate_lib_config_v_1_0_0(migrations["lib"])
                if "node" in migrations and migrations["node"]:
                    LOGGER.info("----------- Migrating Node settings to PyLav 1.0.0 ---------")
                    await migrate_node_config_v_1_0_0(migrations["node"])



from __future__ import annotations

import asyncio
import contextlib
from pathlib import Path
from typing import Any

import discord
from redbot.core.i18n import Translator
from redbot.vendored.discord.ext import menus

from pylav.core.context import PyLavContext
from pylav.extension.red.ui.buttons.generic import CloseButton, NavigateButton, NoButton, RefreshButton, YesButton
from pylav.extension.red.ui.selectors.generic import EntrySelectSelector
from pylav.extension.red.ui.sources.generic import EntryPickerSource
from pylav.helpers.format.strings import shorten_string
from pylav.logging import getLogger
from pylav.type_hints.bot import DISCORD_BOT_TYPE, DISCORD_COG_TYPE, DISCORD_CONTEXT_TYPE, DISCORD_INTERACTION_TYPE

LOGGER = getLogger("PyLav.ext.red.ui.menu.generic")
_ = Translator("PyLav", Path(__file__))


class BaseMenu(discord.ui.View):
    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        bot: DISCORD_BOT_TYPE,
        source: menus.ListPageSource,
        *,
        delete_after_timeout: bool = True,
        timeout: int = 120,
        message: discord.Message = None,
        starting_page: int = 0,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            timeout=timeout,
        )
        self.author = None
        self.ctx = None
        self.cog = cog
        self.bot = bot
        self.message = message
        self._source = source
        self.delete_after_timeout = delete_after_timeout
        self.current_page = starting_page or kwargs.get("page_start", 0)
        self._running = True

    @property
    def source(self) -> menus.ListPageSource:
        return self._source

    async def on_timeout(self):
        self._running = False
        if self.message is None:
            return
        with contextlib.suppress(discord.HTTPException):
            if self.delete_after_timeout and not self.message.flags.ephemeral:
                await self.message.delete()
            else:
                await self.message.edit(view=None)

    async def get_page(self, page_num: int):
        try:
            if page_num >= self._source.get_max_pages():
                page_num = 0
                self.current_page = 0
            page = await self.source.get_page(page_num)
        except IndexError:
            self.current_page = 0
            page = await self.source.get_page(self.current_page)
        value = await self.source.format_page(self, page)
        if isinstance(value, dict):
            return value
        elif isinstance(value, str):
            return {"content": value, "embed": None}
        elif isinstance(value, discord.Embed):
            return {"embed": value, "content": None}

    async def send_initial_message(self, ctx: PyLavContext | DISCORD_INTERACTION_TYPE):
        self.ctx = ctx
        kwargs = await self.get_page(self.current_page)
        await self.prepare()
        self.message = await ctx.send(view=self, ephemeral=True, **kwargs)
        return self.message

    async def show_page(self, page_number, interaction: DISCORD_INTERACTION_TYPE):
        self.current_page = page_number
        kwargs = await self.get_page(self.current_page)
        await self.prepare()
        attachments = []
        if "file" in kwargs:
            attachments = [kwargs.pop("file")]
        elif "files" in kwargs:
            attachments = kwargs.pop("files")
        if attachments:
            kwargs["attachments"] = attachments
        if self.message is not None:
            if not interaction.response.is_done():
                await interaction.response.pong()
            await self.message.edit(view=self, **kwargs)
        elif not interaction.response.is_done():
            await interaction.response.edit_message(view=self, **kwargs)

    async def show_checked_page(self, page_number: int, interaction: DISCORD_INTERACTION_TYPE) -> None:
        max_pages = self._source.get_max_pages()
        with contextlib.suppress(IndexError):
            if max_pages is None or max_pages > page_number >= 0:
                # If it doesn't give maximum pages, it cannot be checked
                await self.show_page(page_number, interaction)
            elif page_number >= max_pages:
                await self.show_page(0, interaction)
            else:
                await self.show_page(max_pages - 1, interaction)

    async def interaction_check(self, interaction: DISCORD_INTERACTION_TYPE):
        """Just extends the default reaction_check to use owner_ids"""
        if (not await self.bot.allowed_by_whitelist_blacklist(interaction.user, guild=interaction.guild)) or (
            self.author and (interaction.user.id != self.author.id)
        ):
            await interaction.response.send_message(
                content=_("You are not authorized to interact with this."), ephemeral=True
            )
            return False
        return True

    async def prepare(self):
        return

    async def on_error(
        self, error: Exception, item: discord.ui.Item[Any], interaction: DISCORD_INTERACTION_TYPE
    ) -> None:
        LOGGER.info("Ignoring exception in view %s for item %s:", self, item, exc_info=error)


class PaginatingMenu(BaseMenu):
    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        bot: DISCORD_BOT_TYPE,
        source: Any,
        original_author: discord.abc.User,
        *,
        clear_buttons_after: bool = True,
        delete_after_timeout: bool = False,
        timeout: int = 120,
        message: discord.Message = None,
        starting_page: int = 0,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            cog=cog,
            bot=bot,
            source=source,
            clear_buttons_after=clear_buttons_after,
            delete_after_timeout=delete_after_timeout,
            timeout=timeout,
            message=message,
            starting_page=starting_page,
            **kwargs,
        )
        self.author = original_author
        self.forward_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK RIGHT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}",
            direction=1,
            row=0,
            cog=cog,
        )
        self.backward_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK LEFT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}",
            direction=-1,
            row=0,
            cog=cog,
        )
        self.first_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK LEFT-POINTING DOUBLE TRIANGLE}",
            direction=0,
            row=0,
            cog=cog,
        )
        self.last_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE}",
            direction=self.source.get_max_pages,
            row=0,
            cog=cog,
        )
        self.refresh_button = RefreshButton(
            style=discord.ButtonStyle.grey,
            row=0,
            cog=cog,
        )

        self.close_button = CloseButton(
            style=discord.ButtonStyle.red,
            row=0,
            cog=cog,
        )
        self.add_item(self.close_button)
        self.add_item(self.first_button)
        self.add_item(self.backward_button)
        self.add_item(self.forward_button)
        self.add_item(self.last_button)

    async def start(self, ctx: PyLavContext | DISCORD_INTERACTION_TYPE):
        if isinstance(ctx, discord.Interaction):
            ctx = await self.cog.bot.get_context(ctx)
        if ctx.interaction and not ctx.interaction.response.is_done():
            await ctx.defer(ephemeral=True)
        self.ctx = ctx
        await self.send_initial_message(ctx)

    async def prepare(self):
        max_pages = self.source.get_max_pages()
        self.forward_button.disabled = False
        self.backward_button.disabled = False
        self.first_button.disabled = False
        self.last_button.disabled = False
        if max_pages == 1:
            self.forward_button.disabled = True
            self.backward_button.disabled = True
            self.first_button.disabled = True
            self.last_button.disabled = True
        elif max_pages == 2:
            self.first_button.disabled = True
            self.last_button.disabled = True


class PromptYesOrNo(discord.ui.View):
    ctx: DISCORD_CONTEXT_TYPE
    message: discord.Message
    author: discord.abc.User
    response: bool

    def __init__(self, cog: DISCORD_COG_TYPE, initial_message: str, *, timeout: int = 120) -> None:
        super().__init__(timeout=timeout)
        self.cog = cog
        self.initial_message_str = initial_message
        self.yes_button = YesButton(
            style=discord.ButtonStyle.green,
            row=0,
            cog=cog,
        )
        self.no_button = NoButton(
            style=discord.ButtonStyle.red,
            row=0,
            cog=cog,
        )
        self.add_item(self.yes_button)
        self.add_item(self.no_button)
        self._running = True
        self.message = None  # type: ignore
        self.ctx = None  # type: ignore
        self.author = None  # type: ignore
        self.response = None  # type: ignore

    async def on_timeout(self):
        self._running = False
        if self.message is None:
            return
        with contextlib.suppress(discord.HTTPException):
            if not self.message.flags.ephemeral:
                await self.message.delete()
            else:
                await self.message.edit(view=None)

    async def start(self, ctx: PyLavContext | DISCORD_INTERACTION_TYPE):
        if isinstance(ctx, discord.Interaction):
            ctx = await self.cog.bot.get_context(ctx)
        if ctx.interaction and not ctx.interaction.response.is_done():
            await ctx.defer(ephemeral=True)
        self.ctx = ctx
        await self.send_initial_message(ctx)

    async def send_initial_message(self, ctx: PyLavContext | DISCORD_INTERACTION_TYPE):
        self.author = ctx.user if isinstance(ctx, discord.Interaction) else ctx.author
        self.ctx = ctx
        self.message = await ctx.send(
            embed=await self.cog.pylav.construct_embed(description=self.initial_message_str, messageable=ctx),
            view=self,
            ephemeral=True,
        )
        return self.message

    async def wait_for_yes_no(self, wait_for: float = None) -> bool:
        tasks = [asyncio.create_task(c) for c in [self.yes_button.responded.wait(), self.no_button.responded.wait()]]
        done, pending = await asyncio.wait(tasks, timeout=wait_for or self.timeout, return_when=asyncio.FIRST_COMPLETED)
        self.stop()
        for task in pending:
            task.cancel()
        if done:
            done.pop().result()
        if not self.message.flags.ephemeral:
            await self.message.delete()
        else:
            await self.message.edit(view=None)
        self.response = bool(self.yes_button.responded.is_set())
        return self.response

    def stop(self):
        super().stop()
        asyncio.ensure_future(self.on_timeout())


class EntryPickerMenu(BaseMenu):
    _source: EntryPickerSource
    result: Any

    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        bot: DISCORD_BOT_TYPE,
        source: EntryPickerSource,
        selector_text: str,
        selector_cls: type[EntrySelectSelector],  # noqa
        original_author: discord.abc.User,
        *,
        clear_buttons_after: bool = False,
        delete_after_timeout: bool = True,
        timeout: int = 120,
        message: discord.Message = None,
        starting_page: int = 0,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            cog,
            bot,
            source,
            clear_buttons_after=clear_buttons_after,
            delete_after_timeout=delete_after_timeout,
            timeout=timeout,
            message=message,
            starting_page=starting_page,
            **kwargs,
        )
        self.result: Any = None
        self.selector_cls = selector_cls
        self.selector_text = shorten_string(max_length=100, string=selector_text)
        self.forward_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK RIGHT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}",
            direction=1,
            row=4,
            cog=cog,
        )
        self.backward_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK LEFT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}",
            direction=-1,
            row=4,
            cog=cog,
        )
        self.first_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK LEFT-POINTING DOUBLE TRIANGLE}",
            direction=0,
            row=4,
            cog=cog,
        )
        self.last_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE}",
            direction=self.source.get_max_pages,
            row=4,
            cog=cog,
        )
        self.close_button = CloseButton(
            style=discord.ButtonStyle.red,
            row=4,
            cog=cog,
        )
        self.refresh_button = RefreshButton(
            style=discord.ButtonStyle.grey,
            row=4,
            cog=cog,
        )
        self.select_view: EntrySelectSelector | None = None
        self.author = original_author

    @property
    def source(self) -> EntryPickerSource:
        return self._source

    async def prepare(self):
        self.clear_items()
        max_pages = self.source.get_max_pages()
        self.forward_button.disabled = False
        self.backward_button.disabled = False
        self.first_button.disabled = False
        self.last_button.disabled = False
        if max_pages == 1:
            self.forward_button.disabled = True
            self.backward_button.disabled = True
            self.first_button.disabled = True
            self.last_button.disabled = True
        elif max_pages == 2:
            self.first_button.disabled = True
            self.last_button.disabled = True
        self.add_item(self.close_button)
        self.add_item(self.first_button)
        self.add_item(self.backward_button)
        self.add_item(self.forward_button)
        self.add_item(self.last_button)
        if self.source.select_options:
            options = self.source.select_options
            self.remove_item(self.select_view)
            self.select_view = self.selector_cls(options, self.cog, self.selector_text, self.source.select_mapping)
            self.add_item(self.select_view)
        if self.select_view and not self.source.select_options:
            self.remove_item(self.select_view)
            self.select_view = None

    async def start(self, ctx: PyLavContext | DISCORD_INTERACTION_TYPE):
        if isinstance(ctx, discord.Interaction):
            ctx = await self.cog.bot.get_context(ctx)
        if ctx.interaction and not ctx.interaction.response.is_done():
            await ctx.defer(ephemeral=True)
        self.ctx = ctx
        await self.send_initial_message(ctx)

    async def show_page(self, page_number: int, interaction: DISCORD_INTERACTION_TYPE):
        await self._source.get_page(page_number)
        await self.prepare()
        self.current_page = page_number
        if self.message is not None:
            if not interaction.response.is_done():
                await interaction.response.pong()
            await self.message.edit(view=self)
        elif not interaction.response.is_done():
            await interaction.response.edit_message(view=self)

    async def wait_for_response(self):
        if isinstance(self.select_view, EntrySelectSelector):
            await asyncio.wait_for(self.select_view.responded.wait(), timeout=self.timeout)
            self.result = self.select_view.entry



from __future__ import annotations

import base64
import os

# noinspection PyProtectedMember
from pylav._internals.functions import _get_path, fix
from pylav.constants.node_features import SUPPORTED_SEARCHES
from pylav.constants.specials import _MAPPING, ANIME
from pylav.logging import getLogger

LOGGER = getLogger("PyLav.Environment")

LOCAL_DEBUGGING = os.getenv("PYLAV__DEBUGGING", False)

POSTGRES_HOST = os.getenv("PYLAV__POSTGRES_HOST", os.getenv("PGHOST"))
# noinspection SpellCheckingInspection
POSTGRES_PORT = os.getenv("PYLAV__POSTGRES_PORT", os.getenv("PGPORT"))
# noinspection SpellCheckingInspection
POSTGRES_PASSWORD = os.getenv("PYLAV__POSTGRES_PASSWORD", os.getenv("PGPASSWORD"))
# noinspection SpellCheckingInspection
POSTGRES_USER = os.getenv("PYLAV__POSTGRES_USER", os.getenv("PGUSER"))
# noinspection SpellCheckingInspection
POSTGRES_DATABASE = os.getenv("PYLAV__POSTGRES_DB", os.getenv("PGDATABASE"))
POSTGRES_SOCKET = os.getenv("PYLAV__POSTGRES_SOCKET")
POSTGRES_CONNECTIONS = (
    max(int(envar_value), 4) if (envar_value := os.getenv("PYLAV__POSTGRES_CONNECTIONS", "100")) is not None else None
)

FALLBACK_POSTGREST_HOST = POSTGRES_HOST
if POSTGRES_SOCKET is not None:
    POSTGRES_PORT = None
    POSTGRES_HOST = POSTGRES_SOCKET
JAVA_EXECUTABLE = _get_path(os.getenv("PYLAV__JAVA_EXECUTABLE") or "java")
USE_BUNDLED_EXTERNAL_PYLAV_NODE = bool(int(os.getenv("PYLAV__USE_BUNDLED_EXTERNAL_PYLAV_NODE", "0")))

REDIS_FULL_ADDRESS_RESPONSE_CACHE = os.getenv("PYLAV__REDIS_FULL_ADDRESS_RESPONSE_CACHE")

EXTERNAL_UNMANAGED_HOST = os.getenv("PYLAV__EXTERNAL_UNMANAGED_HOST")
EXTERNAL_UNMANAGED_PORT = int(os.getenv("PYLAV__EXTERNAL_UNMANAGED_PORT", "80"))
EXTERNAL_UNMANAGED_PASSWORD = os.getenv("PYLAV__EXTERNAL_UNMANAGED_PASSWORD")
EXTERNAL_UNMANAGED_SSL = bool(int(os.getenv("PYLAV__EXTERNAL_UNMANAGED_SSL", 0)))
EXTERNAL_UNMANAGED_NAME = os.getenv("PYLAV__EXTERNAL_UNMANAGED_NAME") or "ENVAR Node (Unmanaged)"

READ_CACHING_ENABLED = bool(int(os.getenv("PYLAV__READ_CACHING_ENABLED", "0")))

TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS = max(
    int(os.getenv("PYLAV__TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS", "1")), 1
)
TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS = max(
    int(os.getenv("PYLAV__TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS", "7")), 7
)
TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS = max(
    int(os.getenv("PYLAV__TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS", "7")), 7
)

# noinspection SpellCheckingInspection
DEFAULT_SEARCH_SOURCE = os.getenv("PYLAV__DEFAULT_SEARCH_SOURCE", "dzsearch")
if DEFAULT_SEARCH_SOURCE not in SUPPORTED_SEARCHES:
    # noinspection SpellCheckingInspection
    LOGGER.warning("Invalid search source %s, defaulting to dzsearch", DEFAULT_SEARCH_SOURCE)
    LOGGER.info("Valid search sources are %s", ", ".join(SUPPORTED_SEARCHES.keys()))
    # noinspection SpellCheckingInspection
    DEFAULT_SEARCH_SOURCE = "dzsearch"

MANAGED_NODE_SPOTIFY_CLIENT_ID = os.getenv("PYLAV__MANAGED_NODE_SPOTIFY_CLIENT_ID", "")
MANAGED_NODE_SPOTIFY_CLIENT_SECRET = os.getenv("PYLAV__MANAGED_NODE_SPOTIFY_CLIENT_SECRET", "")
MANAGED_NODE_SPOTIFY_COUNTRY_CODE = os.getenv("PYLAV__MANAGED_NODE_SPOTIFY_COUNTRY_CODE", "US")
MANAGED_NODE_APPLE_MUSIC_API_KEY = os.getenv("PYLAV__MANAGED_NODE_APPLE_MUSIC_API_KEY", "")
MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE = os.getenv("PYLAV__MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE", "US")
MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN = os.getenv("PYLAV__MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN", "")
MANAGED_NODE_DEEZER_KEY = os.getenv("PYLAV__MANAGED_NODE_DEEZER_KEY") or ANIME
if MANAGED_NODE_DEEZER_KEY and MANAGED_NODE_DEEZER_KEY.startswith("id"):
    _temp = [MANAGED_NODE_DEEZER_KEY[i : i + 16] for i in range(0, len(MANAGED_NODE_DEEZER_KEY), 16)]
    MANAGED_NODE_DEEZER_KEY = "".join(
        [
            base64.b64decode(r).decode()
            for r in [
                fix(_temp[2], _MAPPING[2]),
                fix(_temp[1], _MAPPING[1]),
                fix(_temp[3], _MAPPING[3]),
                fix(_temp[0], _MAPPING[0]),
            ]
        ]
    )
LOCAL_TRACKS_FOLDER = os.getenv("PYLAV__LOCAL_TRACKS_FOLDER")
DATA_FOLDER = os.getenv("PYLAV__DATA_FOLDER")
ENABLE_NODE_RESUMING = bool(int(os.getenv("PYLAV__ENABLE_NODE_RESUMING", "1")))
DEFAULT_PLAYER_VOLUME = (
    max(int(envar_value), 1) if (envar_value := os.getenv("PYLAV__DEFAULT_PLAYER_VOLUME")) is not None else None
)



from __future__ import annotations

import dataclasses
from typing import Literal, TypeAlias, Union

from pylav.exceptions.request import HTTPException
from pylav.nodes.api.responses.exceptions import LoadException
from pylav.nodes.api.responses.filters import Filters
from pylav.nodes.api.responses.misc import Git, Plugin, Version
from pylav.nodes.api.responses.player import State
from pylav.nodes.api.responses.playlists import Info
from pylav.nodes.api.responses.shared import PlaylistPluginInfo
from pylav.nodes.api.responses.track import Track
from pylav.nodes.api.responses.websocket import CPU, Frame, Memory
from pylav.type_hints.dict_typing import JSON_DICT_TYPE


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class PlaylistData:
    info: Info
    pluginInfo: PlaylistPluginInfo
    tracks: list[Track]

    def __post_init__(self):
        temp = []
        for s in self.tracks:
            if isinstance(s, Track) or (isinstance(s, dict) and (s := Track(**s))):
                temp.append(s)
        object.__setattr__(self, "tracks", temp)


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class BaseTrackResponse:
    loadType: Literal["track", "playlist", "search", "empty", "error"]
    data: Track | PlaylistData | list[Track] | LoadException | None

    def __bool__(self):
        return True


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class TrackResponse(BaseTrackResponse):
    loadType: Literal["track"]
    data: Track


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class PlaylistResponse(BaseTrackResponse):
    loadType: Literal["playlist"]
    data: PlaylistData


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class SearchResponse(BaseTrackResponse):
    loadType: Literal["search"]
    data: list[Track]

    def __post_init__(self):
        temp = []
        for s in self.data:
            if isinstance(s, Track) or (isinstance(s, dict) and (s := Track(**s))):
                temp.append(s)
        object.__setattr__(self, "data", temp)


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class EmptyResponse(BaseTrackResponse):
    loadType: Literal["empty"]
    data: None = None

    def __bool__(self):
        return False


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class ErrorResponse(BaseTrackResponse):  # noqa
    loadType: Literal["error"]
    data: LoadException

    def __bool__(self):
        return False


LoadTrackResponses: TypeAlias = Union[
    TrackResponse, PlaylistResponse, EmptyResponse, ErrorResponse, SearchResponse, HTTPException
]


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class LavalinkInfo:
    version: Version
    buildTime: int
    git: Git
    jvm: str
    lavaplayer: str
    sourceManagers: list[str]
    filters: list[str]
    plugins: list[Plugin]

    def __post_init__(self):
        temp = []
        for s in self.plugins:
            if isinstance(s, Plugin) or (isinstance(s, dict) and (s := Plugin(**s))):
                temp.append(s)
        object.__setattr__(self, "plugins", temp)


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class VoiceState:
    token: str
    endpoint: str
    sessionId: str

    def to_dict(self) -> JSON_DICT_TYPE:
        return {
            "token": self.token,
            "endpoint": self.endpoint,
            "sessionId": self.sessionId,
        }

    def __repr__(self) -> str:
        return (
            f"<VoiceStateObject(token={'OBFUSCATED' if self.token else None} "
            f"endpoint={self.endpoint} "
            f"sessionId={self.sessionId})"
        )


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class LavalinkPlayer:
    guildId: str
    volume: int
    paused: bool
    state: State | dict
    voice: VoiceState | dict
    filters: Filters | dict
    track: Track | dict | None = None

    def __post_init__(self) -> None:
        if isinstance(self.voice, dict):
            object.__setattr__(self, "voice", VoiceState(**self.voice))
        if isinstance(self.filters, dict):
            object.__setattr__(self, "filters", Filters(**self.filters))
        if isinstance(self.track, dict):
            object.__setattr__(self, "track", Track(**self.track))
        if isinstance(self.state, dict):
            object.__setattr__(self, "state", State(**self.state))

    def to_dict(
        self,
    ) -> JSON_DICT_TYPE:
        assert (
            isinstance(self.voice, VoiceState)
            and isinstance(self.filters, Filters)
            and isinstance(self.track, (Track, type(None)))
        )
        return {
            "guildId": self.guildId,
            "volume": self.volume,
            "paused": self.paused,
            "state": self.state.to_dict(),
            "voice": self.voice.to_dict(),
            "filters": self.filters.to_dict(),
            "track": self.track.to_dict() if self.track else None,
        }


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Stats:
    players: int
    playingPlayers: int
    uptime: int
    memory: Memory
    cpu: CPU
    frameStats: Frame | None = None
    uptime_seconds: int = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        object.__setattr__(self, "uptime_seconds", self.uptime / 1000)



from __future__ import annotations

import collections
from typing import Any, Final

from deepdiff import DeepDiff  # type: ignore

from pylav.players.filters.misc import FilterMixin

_SUPPORTED_BANDS: Final = 15  # 1 Indexed


class Equalizer(FilterMixin):
    """Class representing a usable equalizer.

    Parameters
    ------------
    levels: List[Tuple[int, float]]
        A list of tuple pairs containing a band int and gain float.
    name: str
        An Optional string to name this Equalizer. Defaults to 'CustomEqualizer'
    """

    __slots__ = ("_eq", "_name", "_default")

    def __init__(self, *, levels: list[dict[str, int | float | None]], name: str = "CustomEqualizer") -> None:
        super().__init__()
        self._eq = self._factory(levels)
        self._name = name

    def to_dict(self) -> dict[str, list[dict[str, int | float | None]] | str | bool]:
        """Returns a dictionary representation of the Equalizer"""
        return {"equalizer": self._eq, "name": self._name}

    @classmethod
    def from_dict(cls, data: dict[str, list[dict[str, int | float | None]] | str | bool]) -> Equalizer:
        """Creates an Equalizer from a dictionary"""
        return cls(levels=data["equalizer"], name=data["name"])

    def __str__(self) -> str:
        return self._name

    def __repr__(self) -> str:
        return f"<Equalizer: name={self._name}, eq={self._eq}>"

    @property
    def index(self) -> dict[int, float]:
        d: dict[Any, float] = collections.defaultdict(float)
        d |= {d["band"]: d["gain"] for d in self._eq}
        return d

    def __eq__(self, other: Any) -> bool:
        """Overrides the default implementation"""
        if isinstance(other, Equalizer):
            return bool(
                DeepDiff(
                    self._eq, other._eq, ignore_order=True, max_passes=1, cache_size=100, exclude_paths=["root['name']"]
                )
            )
        return NotImplemented

    @property
    def name(self) -> str:
        """The Equalizers friendly name"""
        return self._name

    @staticmethod
    def _factory(levels: list[dict[str, int | float]]) -> list[dict[str, int | float]]:
        if not levels:
            return []
        if not isinstance(levels[0], dict):
            raise TypeError("Equalizer levels should be a list of dictionaries")

        _dict: dict[str, float] = collections.defaultdict(float)
        for level in levels:
            _dict[str(level["band"])] = level["gain"]
        return [{"band": i, "gain": _dict[str(i)]} for i in range(_SUPPORTED_BANDS) if _dict[str(i)]]

    @classmethod
    def build(cls, *, levels: list[dict[str, int | float]], name: str = "CustomEqualizer") -> Equalizer:
        """Build a custom Equalizer class with the provided levels.

        Parameters
        ------------
        levels: list[dict[str, int | float]]
            A list of dictionaries containing the band and gain for each band.
        name: str
            An Optional string to name this Equalizer. Defaults to 'CustomEqualizer'
        """
        return cls(levels=levels, name=name)

    @classmethod
    def flat(cls) -> Equalizer:
        """Flat Equalizer.
        Resets your EQ to Flat.
        """
        return cls(
            levels=[],
            name="Default",
        )

    @classmethod
    def default(cls) -> Equalizer:
        return cls.flat()

    def get(self) -> list[dict[str, int | float]]:
        return [] if self.off else self._eq

    def reset(self) -> None:
        eq = Equalizer.flat()
        self._eq = eq._eq
        self._name = eq._name

    def set_gain(self, band: int, gain: float) -> None:
        if band < 0 or band >= _SUPPORTED_BANDS:
            raise IndexError(f"Band {band} does not exist!")

        band = next((index for (index, d) in enumerate(self._eq) if d["band"] == band), -1)
        if band == -1:
            raise IndexError(f"Band {band} does not exist!")

        gain = float(min(max(gain, -0.25), 1.0))
        self._eq[band]["gain"] = gain
        if gain == 0.0:
            # Discard any redundant 0.0 gains
            self._eq[band].pop("gain", 0.0)

    def get_gain(self, band: int) -> float:
        if band < 0 or band >= _SUPPORTED_BANDS:
            raise IndexError(f"Band {band} does not exist!")
        return self.index[band]

    def visualise(self) -> str:
        block = ""
        bands = [str(band + 1).zfill(2) for band in range(_SUPPORTED_BANDS)]
        bottom = (" " * 8) + " ".join(bands)
        gains = [x * 0.01 for x in range(-25, 105, 5)]
        gains.reverse()

        for gain in gains:
            prefix = ""
            if gain > 0:
                prefix = "+"
            elif gain == 0:
                prefix = " "

            block += f"{prefix}{gain:.2f} | "

            for value in self._eq:
                cur_gain = value.get("gain", 0.0)
                block += "[] " if cur_gain >= gain else "   "
            block += "\n"

        block += bottom
        return block



from __future__ import annotations

from typing import TYPE_CHECKING

from pylav.events.base import PyLavEvent
from pylav.nodes.api.responses.websocket import TrackStart

if TYPE_CHECKING:
    from pylav.nodes.node import Node
    from pylav.players.player import Player
    from pylav.players.tracks.obj import Track


class TrackStartEvent(PyLavEvent):
    """This event is dispatched when the player starts to play a track.

    Event can be listened to by adding a listener with the name `pylav_track_start_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.
    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        self.player = player
        self.track = track
        self.node = node
        self.event = event_object


class TrackStartYouTubeEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a YouTube track.

    Event can be listened to by adding a listener with the name `pylav_track_start_youtube_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


# noinspection SpellCheckingInspection
class TrackStartClypitEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Clyp.it track.

    Event can be listened to by adding a listener with the name `pylav_track_start_clypit_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartGetYarnEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a GetYarn track.

    Event can be listened to by adding a listener with the name `pylav_track_start_get_yarn_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartMixCloudEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a MixCloud track.

    Event can be listened to by adding a listener with the name `pylav_track_start_mix_cloud_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartOCRMixEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a OCR Mix track.

    Event can be listened to by adding a listener with the name `pylav_track_start_ocr_mix_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartPornHubEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Pornhub track.

    Event can be listened to by adding a listener with the name `pylav_track_start_porn_hub_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartRedditEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Reddit track.

    Event can be listened to by adding a listener with the name `pylav_track_start_reddit_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


# noinspection SpellCheckingInspection
class TrackStartSoundgasmEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Soundgasm track.

    Event can be listened to by adding a listener with the name `pylav_track_start_soundgasm_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartTikTokEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a TikTok track.

    Event can be listened to by adding a listener with the name `pylav_track_start_tik_tok_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartSpotifyEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Spotify track.

    Event can be listened to by adding a listener with the name `pylav_track_start_spotify_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartDeezerEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Deezer track.

    Event can be listened to by adding a listener with the name `pylav_track_start_deezer_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartYandexMusicEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Yandex Music track.

    Event can be listened to by adding a listener with the name `pylav_track_start_yandex_music_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartAppleMusicEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play an Apple Music track.

    Event can be listened to by adding a listener with the name `pylav_track_start_apple_music_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartBandcampEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Bandcamp track.

    Event can be listened to by adding a listener with the name `pylav_track_start_bandcamp_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartYouTubeMusicEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a YouTube Music track.

    Event can be listened to by adding a listener with the name `pylav_track_start_youtube_music_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartSoundCloudEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a SoundCloud track.

    Event can be listened to by adding a listener with the name `pylav_track_start_soundcloud_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartTwitchEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Twitch track.

    Event can be listened to by adding a listener with the name `pylav_track_start_twitch_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartHTTPEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play an HTTP track.

    Event can be listened to by adding a listener with the name `pylav_track_start_http_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartLocalFileEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a local file track.

    Event can be listened to by adding a listener with the name `pylav_track_start_local_file_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartNicoNicoEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a NicoNico track.

    Event can be listened to by adding a listener with the name `pylav_track_start_nico_nico_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartVimeoEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Vimeo track.

    Event can be listened to by adding a listener with the name `pylav_track_start_vimeo_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartSpeakEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Speak track.

    Event can be listened to by adding a listener with the name `pylav_track_start_speak_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)


class TrackStartGCTTSEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Google Cloud TTS track.

    Event can be listened to by adding a listener with the name `pylav_track_start_gctts_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play the track.
    track: :class:`Track`
        The track that was started.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStart`
        The raw event object.

    """

    __slots__ = ("player", "track", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        super().__init__(player, track, node, event_object)



from __future__ import annotations

from datetime import datetime

import pytz


def get_now_utc() -> datetime:
    """A helper function to return an aware UTC datetime representing the current time."""
    return datetime.now(tz=get_tz_utc())


def get_tz_utc() -> pytz.UTC:
    """A helper function to return a pytz.UTC object."""
    return pytz.UTC



from __future__ import annotations

from piccolo.columns import BigInt, Text
from piccolo.table import Table

from pylav.constants.versions import VERSION_0_0_0
from pylav.storage.database.tables.misc import DATABASE_ENGINE


class BotVersionRow(Table, db=DATABASE_ENGINE, tablename="version"):
    bot = BigInt(primary_key=True, index=True)
    version = Text(null=False, default=str(VERSION_0_0_0))



from __future__ import annotations

import discord

from pylav.core.context import PyLavContext
from pylav.logging import getLogger
from pylav.type_hints.bot import DISCORD_BOT_TYPE, DISCORD_INTERACTION_TYPE

RED_LOGGER = getLogger("red")
LOGGER = getLogger("PyLav.bot")


async def process_commands(self, message: discord.Message, /):
    """
    Same as base method, but dispatches an additional event for cogs
    which want to handle normal messages differently to command
    messages,  without the overhead of additional get_context calls
    per cog.
    """
    if not message.author.bot:
        ctx = await self.get_context(message, cls=PyLavContext)
        if ctx.invoked_with and isinstance(message.channel, discord.PartialMessageable):
            RED_LOGGER.warning(
                "Discarded a command message (ID: %s) with PartialMessageable channel: %r",
                message.id,
                message.channel,
            )
        else:
            await self.invoke(ctx)
    else:
        ctx = None

    if ctx is None or ctx.valid is False:
        self.dispatch("message_without_command", message)


async def get_context(
    self: DISCORD_BOT_TYPE, message: discord.Message | DISCORD_INTERACTION_TYPE, /, *, cls=PyLavContext
) -> PyLavContext:
    """Get the context for a command invocation."""
    return await super(self.__class__, self).get_context(message, cls=cls)  # noqa



from __future__ import annotations

__all__ = (
    "SUPPORTED_SEARCHES",
    "SUPPORTED_SOURCES",
    "SUPPORTED_FEATURES",
    "SUPPORTED_FILTERS",
)

# noinspection SpellCheckingInspection
SUPPORTED_SEARCHES = {
    "ytmsearch": "YouTube Music",
    "ytsearch": "YouTube",
    "spsearch": "Spotify",
    "scsearch": "SoundCloud",
    "amsearch": "Apple Music",
    "dzsearch": "Deezer",
}

# noinspection SpellCheckingInspection
SUPPORTED_SOURCES = {
    # https://github.com/lavalink-devs/Lavalink
    "youtube",
    "soundcloud",
    "bandcamp",
    "twitch",
    "vimeo",
    "local",
    "http",
    # https://github.com/DuncteBot/skybot-lavalink-plugin
    "getyarn.io",
    "clypit",
    "speak",
    "pornhub",
    "reddit",
    "ocremix",
    "tiktok",
    "mixcloud",
    "soundgasm",
    # https://github.com/TopiSenpai/LavaSrc
    "spotify",
    "applemusic",
    "deezer",
    "yandexmusic",
    # https://github.com/DuncteBot/tts-plugin
    "gcloud-tts",
}

SUPPORTED_FEATURES = {
    # https://github.com/Topis-Lavalink-Plugins/Sponsorblock-Plugin
    "sponsorblock",
}
SUPPORTED_FILTERS = {
    "distortion",
    "volume",
    "karaoke",
    "echo",
    "equalizer",
    "timescale",
    "tremolo",
    "lowPass",
    "rotation",
    "channelMix",
    "vibrato",
}
